import 'dart:convert';

import 'package:dio/dio.dart';
import 'package:memri/constants/app_settings.dart';
import 'package:memri/core/apis/base_api.dart';

class GitlabAPI extends BaseAPI {
  GitlabAPI() : super(AppSettings().gitlabBaseUrl, isPodUrl: false);

  String? gitlabAccessToken;
  late Function refreshCallback;

  /// See https://docs.gitlab.com/ee/api/repository_files.html#get-file-from-repository
  /// for the documentation of the endpoint.
  Future<String> getTextFile({
    int? projectId,
    String? projectPath,
    required String filepath,
    String branch = 'main',
  }) async {
    if (projectId == null && projectPath == null) {
      throw ArgumentError('Either projectId or projectName must be provided');
    }
    final project = projectId ?? Uri.encodeComponent(projectPath!);
    final encodedPath = Uri.encodeComponent(filepath);
    final endpoint =
        '$baseUrl/projects/$project/repository/files/$encodedPath?ref=$branch';
    final response = await dio.get(endpoint);
    checkResponseError(response);
    final decodedGitJson = response.data;
    if (decodedGitJson.length == 0 || decodedGitJson['content'] == null) {
      throw '$filepath is not available in provided repository link';
    }
    return Utf8Decoder().convert(base64Decode(decodedGitJson['content']));
  }

  Future<String> downloadSingleArtifact({
    required int gitProjectId,
    required String filename,
    required String jobName,
    String? branch = 'main',
  }) async {
    final endpoint =
        '$baseUrl/projects/$gitProjectId/jobs/artifacts/$branch/raw/$filename?job=$jobName';
    final response = await dio.get(endpoint);
    checkResponseError(response);
    return response.data;
  }

  Map<String, String> get headers =>
      {'Authorization': 'bearer $gitlabAccessToken'};

  Future<Response<T>> get<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onReceiveProgress,
  }) async {
    Future<Response<T>> call() async {
      return await dio.get(path,
          queryParameters: queryParameters,
          options: options,
          cancelToken: cancelToken,
          onReceiveProgress: onReceiveProgress);
    }

    try {
      return call();
    } catch (e) {
      refreshCallback();
      return call();
    }
  }

  Future<Map<String, dynamic>> getUser() async {
    String endpoint = '$baseUrl/user';
    final response = await get(endpoint, options: Options(headers: headers));
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> getUserProjects(userId) async {
    String endpoint = '$baseUrl/users/$userId/projects';
    final response = await get(endpoint, options: Options(headers: headers));
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> getProjectPackages(int projectId) async {
    String endpoint = '$baseUrl/projects/$projectId/packages';
    final response = await get(endpoint, options: Options(headers: headers));
    checkResponseError(response);
    return response.data;
  }

  Future<int?> getLastPipelineId(dynamic projectId) async {
    String endpoint = '$baseUrl/projects/$projectId/pipelines?per_page=1';
    final response = await get(endpoint, options: Options(headers: headers));
    checkResponseError(response);
    return response.data[0]['id'];
  }

  Future<dynamic> getJobsJson(dynamic projectId, dynamic pipelineId) async {
    String endpoint = '$baseUrl/projects/$projectId/pipelines/$pipelineId/jobs';
    final response = await get(endpoint, options: Options(headers: headers));
    checkResponseError(response);
    return response.data;
  }

// @Deprecated('Old Implementation')
// Future<String> downloadSingleArtifact(
//     {required int gitProjectId,
//       required String filename,
//       required String jobName,
//       String branch = 'main'}) async {
//   final repoUri = Uri.parse(
//       'https://gitlab.memri.io/api/v4/projects/$gitProjectId/jobs/artifacts/$branch/raw/$filename?job=$jobName');
//   final response =
//   await http.get(repoUri, headers: {'content-type': 'application/json'});
//   if (response.statusCode != 200) {
//     throw 'ERROR: ${response.statusCode} ${response.reasonPhrase}';
//   }
//   return Utf8Decoder().convert(response.bodyBytes);
// }
}
