import 'dart:convert';

import 'package:dio/dio.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/apis/base_api.dart';
import 'package:memri/core/models/pod/pod_config.dart';
import 'package:moor/moor.dart';

class PodAPI extends BaseAPI {
  PodAPI() : super('');

  late PodConfig _podConfig;
  String _endpointUrl = '';

  void setConnectionConfig(PodConfig podConfig) {
    _podConfig = podConfig;
    _endpointUrl =
        '${_podConfig.baseUrl}/${_podConfig.apiVersion}/${_podConfig.ownerKey}';
  }

  Future<dynamic> authenticate() async {
    String endpoint = '$_endpointUrl/search';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {'_limit': 1},
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> search(dynamic payload) async {
    var endpoint = '$_endpointUrl/search';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': payload,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<String> createItem(Map<String, dynamic> item_map) async {
    String endpoint = '$_endpointUrl/create_item';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': item_map,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> updateItem(Map<String, dynamic> syncDict) async {
    String endpoint = '$_endpointUrl/update_item';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': syncDict,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> deleteItem(String itemId) async {
    String endpoint = '$_endpointUrl/delete_item';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': itemId,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<Map<String, dynamic>> getItem(String id) async {
    String endpoint = '$_endpointUrl/get_item';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': id,
      },
    );
    checkResponseError(response);

    if (response.data.isEmpty && response.data is List) {
      var error = 'ERROR 400 - Item $id does not exist';
      AppLogger.err(error);
      throw error;
    }

    var res = (response.data as List)[0];
    return res;
  }

  Future<dynamic> getLogsForPluginRun(String itemId) async {
    String endpoint = '$_endpointUrl/get_pluginrun_log';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': itemId,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> bulkAction(dynamic payload) async {
    String endpoint = '$_endpointUrl/bulk';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': payload,
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<String> podVersion() async {
    String endpoint = '$baseUrl/version';
    var response = await dio.get(endpoint);
    checkResponseError(response);
    return jsonDecode(response.data)['cargo'];
  }

  Future<void> deleteEdge(Map<String, String> payload) async {
    String endpoint = '$_endpointUrl/delete_edge_by_source_target';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': payload,
      },
    );
    checkResponseError(response);
  }

  Future<Uint8List> downloadFile(String fileSHAHash) async {
    String endpoint = '$_endpointUrl/get_file';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {'sha256': fileSHAHash},
      },
      options: Options(responseType: ResponseType.bytes),
    );
    checkResponseError(response);
    return response.data;
  }

  /// OAUTH

  Future<dynamic> oauthAccessToken(
      String oauthVerifier, String oauthToken, String oauthTokenSecret) async {
    String endpoint = '$_endpointUrl/oauth1_access_token';
    AppLogger.info('requesting oauth access token');
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {
          'oauthVerifier': oauthVerifier,
          'oauthToken': oauthToken,
          'oauthTokenSecret': oauthTokenSecret
        },
      },
    );
    checkResponseError(response);
    return response.data;
  }

  Future<dynamic> oauthRequestToken(String service, String callback_url) async {
    String endpoint = '$_endpointUrl/oauth1_request_token';
    AppLogger.info('requesting oauth');
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {'service': service, 'callbackUrl': callback_url},
      },
    );

    checkResponseError(response);
    return response.data;
  }

  /// GraphQL

  Future<Map<String, dynamic>> queryGraphQL(String query) async {
    String endpoint = '$_endpointUrl/graphql';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': query,
      },
    );
    checkResponseError(response);
    return jsonDecode(response.toString());
  }

  Future<dynamic> getQrCodeSvg(String webserverUrl) async {
    String endpoint = '$webserverUrl/qr_svg';
    Response response;
    try {
      response = await dio.get(endpoint);
    } on DioError catch (_) {
      return null;
    }
    checkResponseError(response);
    return response.data;
  }

  Future<Map<String, dynamic>> oauth2Authorize(
      String code, String redirectUrl, String platform) async {
    String endpoint = '$_endpointUrl/oauth2/authorize';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {
          'authCode': code,
          'redirectUri': redirectUrl,
          'platform': platform
        }
      },
    );
    checkResponseError(response);
    return jsonDecode(response.toString());
  }

  Future<Map<String, dynamic>> oauth2GetAccessToken(String platform) async {
    String endpoint = '$_endpointUrl/oauth2/access_token';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {'platform': platform}
      },
    );
    checkResponseError(response);
    return jsonDecode(response.toString());
  }

  Future<String> oauth2url(
      String redirectUrl, List<String> scopes, String service) async {
    String endpoint = '$_endpointUrl/oauth2/auth_url';
    var response = await dio.post(
      endpoint,
      data: {
        'auth': {'type': 'ClientAuth', 'databaseKey': _podConfig.databaseKey},
        'payload': {
          'platform': service,
          'redirectUri': redirectUrl,
          'scopes': scopes
        }
      },
    );
    checkResponseError(response);
    return response.toString();
  }
}
