//  Created by T Brennan on 14/12/20.

import 'dart:async';
import 'dart:isolate';

import 'package:flutter/foundation.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/database/auth_key.dart';
import 'package:memri/core/apis/pod_api.dart';
import 'package:memri/core/controllers/database_controller.dart';
import 'package:memri/core/controllers/permission_controller.dart';
import 'package:memri/core/models/database/item_record.dart';
import 'package:memri/core/models/pod/pod_config.dart';
import 'package:memri/core/models/pod_setup.dart';
import 'package:memri/core/services/settings.dart';

enum AppState {
  setup,
  keySaving,
  authentication,
  authenticated,
  incompatibleDevice,
  incompatibleBrowser,
  maintenance
}

enum SystemError {
  connectionLost,
  failedToConnectToPod,
  podNotResponding,
  syncFailed,
  generalError
}

extension SystemErrorExt on SystemError {
  int get weight {
    switch (this) {
      case SystemError.connectionLost:
        return 100;
      case SystemError.generalError:
        return 1000;
      default:
        return 0;
    }
  }

  String get errorString {
    switch (this) {
      case SystemError.connectionLost:
        return 'No internet connection';
      case SystemError.failedToConnectToPod:
        return 'Something went wrong: failed to connect to Pod.';
      case SystemError.podNotResponding:
        return "Pod doesn't respond.";
      case SystemError.syncFailed:
        return 'Failed to synchronize with pod.';
      case SystemError.generalError:
        return 'Sorry! Something went wrong.';
    }
  }

  bool get showDismiss {
    switch (this) {
      case SystemError.syncFailed:
        return true;
      default:
        return false;
    }
  }

  bool get showRetry {
    switch (this) {
      case SystemError.generalError:
        return true;
      default:
        return false;
    }
  }
}

class AppController {
  static AppController shared = AppController();

  late DatabaseController databaseController;
  late PermissionsController permissionController;
  late PodSetupModel model;
  late PodAPI podApi;

  Isolate? syncIsolate;
  final ValueNotifier<AppState> _state = ValueNotifier(AppState.setup);
  PodConfig? _podConnectionConfig;

  bool isDevelopersMode = false;
  final bool _isInDemoMode = false;

  ValueNotifier<AppState> get state => _state;

  set state(newValue) => _state.value = newValue;

  bool get shouldShowError => lastError.value != null;

  showError(SystemError error) {
    if ((lastError.value?.weight ?? 0) > error.weight) return;
    lastError.value = error;
  }

  hideError(SystemError error) {
    if (lastError.value == error) lastError.value = null;
  }

  ValueNotifier<SystemError?> lastError = ValueNotifier(null);

  bool hasNetworkConnection = true;

  //TODO: hope this is temporary solution
  Map<String, dynamic> storage = {};

  AppController() {
    podApi = PodAPI();
    permissionController = PermissionsController();
    model = PodSetupModel();
  }

  Future<AuthKeys> getOwnerAndDBKey() async =>
      await ItemRecord.getOwnerAndDBKey();

  // MARK: Pod connection
  Future<PodConfig?> get podConnectionConfig async {
    if (_isInDemoMode) return null;
    try {
      // Here you should retrieve the connection details stored in the database
      if (_podConnectionConfig == null) {
        var podURL = await Settings.shared.get('defaults/pod/url');
        var keys = await getOwnerAndDBKey();
        var ownerKey = keys.ownerKey;
        var databaseKey = keys.dbKey;
        if (podURL == null) return null;
        var uri = Uri.parse(podURL);
        _podConnectionConfig = PodConfig(
            scheme: uri.scheme,
            host: uri.host,
            port: uri.port,
            ownerKey: ownerKey,
            databaseKey: databaseKey);
      }
      return _podConnectionConfig!;
    } on Exception catch (error, stackTrace) {
      AppLogger.err(error, stackTrace: stackTrace);
      return null;
    }
  }
}
