import 'package:flutter/material.dart';
import 'package:memri/core/cvu/controllers/view_context_controller.dart';
import 'package:memri/utilities/extensions/collection.dart';
import 'package:memri/widgets/empty.dart';

class PhotoViewerView extends StatefulWidget {
  final void Function(bool) onToggleOverlayVisibility;
  final Future<PhotoViewerControllerPhotoItem?> Function(int index)
      photoItemProvider;
  final int initialIndex;
  final ViewContextController viewContext;

  const PhotoViewerView(
      {Key? key,
      required this.onToggleOverlayVisibility,
      required this.viewContext,
      required this.photoItemProvider,
      required this.initialIndex})
      : super(key: key);

  @override
  PhotoViewerViewState createState() => PhotoViewerViewState();
}

class PhotoViewerViewState extends State<PhotoViewerView> {
  late PageController _pageController;

  late final Future<List<PhotoViewerControllerPhotoItem?>> _resolveImages;

  @override
  initState() {
    super.initState();
    _resolveImages = resolveImages();
  }

  Future<List<PhotoViewerControllerPhotoItem?>> resolveImages() async {
    return await Future.wait(
        widget.viewContext.items.mapIndexed((index, item) async {
      return await widget.photoItemProvider(index);
    }));
  }

  @override
  Widget build(BuildContext context) {
    _pageController = PageController(initialPage: widget.initialIndex);
    return FutureBuilder(
        future: _resolveImages,
        builder: (BuildContext context,
            AsyncSnapshot<List<PhotoViewerControllerPhotoItem?>> snapshot) {
          if (snapshot.connectionState == ConnectionState.done) {
            var photoItems = snapshot.data;

            if (photoItems == null || photoItems.isEmpty) {
              return Empty();
            }
            return PageView(
                controller: _pageController, children: getImages(photoItems));
          } else {
            return Center(
              child: SizedBox(
                width: 60,
                height: 60,
                child: CircularProgressIndicator(),
              ),
            );
          }
        });
  }

  List<Widget> getImages(List<PhotoViewerControllerPhotoItem?> photoItems) {
    return photoItems
        .compactMap((photoItem) =>
            photoItem == null ? null : PhotoScalingView(photoItem: photoItem))
        .toList();
  }
//TODO: overlay changing with animations
}

class PhotoViewerControllerPhotoItem {
  final int index;
  final ImageProvider imageProvider;
  final Widget overlay;

  PhotoViewerControllerPhotoItem(this.index, this.imageProvider, this.overlay);
}

class PhotoScalingView extends StatefulWidget {
  final PhotoViewerControllerPhotoItem photoItem;

  const PhotoScalingView({Key? key, required this.photoItem}) : super(key: key);

  @override
  PhotoScalingViewState createState() => PhotoScalingViewState();
}

class PhotoScalingViewState extends State<PhotoScalingView>
    with SingleTickerProviderStateMixin {
  late AnimationController _animationController;

  late Animation<Matrix4> _animation;

  final _transformationController = TransformationController();

  late TapDownDetails _doubleTapDetails;

  @override
  void initState() {
    super.initState();
    _animationController = AnimationController(
      vsync: this,
      duration: Duration(milliseconds: 200),
    )..addListener(() {
        _transformationController.value = _animation.value;
      });
  }

  @override
  void dispose() {
    _animationController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      onDoubleTapDown: _handleDoubleTapDown,
      onDoubleTap: _handleDoubleTap,
      child: InteractiveViewer(
        transformationController: _transformationController,
        child: Image(
          image: widget.photoItem.imageProvider,
        ),
      ),
    );
  }

  void _handleDoubleTapDown(TapDownDetails details) {
    _doubleTapDetails = details;
  }

  void _handleDoubleTap() {
    Matrix4 endMatrix;
    Offset position = _doubleTapDetails.localPosition;

    if (_transformationController.value != Matrix4.identity()) {
      endMatrix = Matrix4.identity();
    } else {
      endMatrix = Matrix4.identity()
        ..translate(-position.dx * 2, -position.dy * 2)
        ..scale(3.0);
    }

    _animation = Matrix4Tween(
      begin: _transformationController.value,
      end: endMatrix,
    ).animate(
      CurveTween(curve: Curves.easeOut).animate(_animationController),
    );
    _animationController.forward(from: 0);
  }
}
