import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:get_it/get_it.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/database/auth_key.dart';
import 'package:memri/core/controllers/app_controller.dart';
import 'package:memri/core/controllers/database_controller.dart';
import 'package:memri/core/models/database/database.dart';
import 'package:memri/core/models/database/item_edge_record.dart';
import 'package:memri/core/models/database/item_property_record.dart';
import 'package:memri/core/services/database/property_database_value.dart';
import 'package:memri/core/services/database/schema.dart';
import 'package:memri/utilities/extensions/collection.dart';
import 'package:moor/moor.dart';
import 'package:uuid/uuid.dart';

part 'item_record.g.dart';

enum SyncState {
  skip,
  create,
  update,
  noChanges,
  failed,
}

enum FileState { skip, needsUpload, needsDownload, noChanges, failedDownload }

extension SyncStateExtension on SyncState {
  static SyncState rawValue(String value) =>
      SyncState.values.firstWhere((val) => val.inString == value);

  String get inString {
    return toString().split('.').last;
  }
}

extension FileStateExtension on FileState {
  static FileState rawValue(String value) =>
      FileState.values.firstWhere((val) => val.inString == value);

  String get inString {
    return toString().split('.').last;
  }
}

@JsonSerializable()
class ItemRecord with EquatableMixin {
  int? rowId;
  String uid;
  String type;
  DateTime dateCreated;
  DateTime? dateServerModified;
  DateTime dateModified;
  bool deleted;

  SyncState syncState;
  FileState fileState;
  bool syncHasPriority;

  ItemRecord(
      {this.rowId,
      String? uid,
      required this.type,
      DateTime? dateCreated,
      DateTime? dateModified,
      this.deleted = false,
      this.syncState = SyncState.create,
      this.fileState = FileState.skip,
      this.syncHasPriority = false})
      : dateModified = dateModified ?? DateTime.now(),
        dateCreated = dateCreated ?? DateTime.now(),
        uid = uid ?? Uuid().v4();

  ItemRecord.fromItem(Item item)
      : rowId = item.rowId,
        uid = item.id,
        type = item.type,
        dateCreated = item.dateCreated,
        dateModified = item.dateModified,
        dateServerModified = item.dateServerModified,
        deleted = item.deleted,
        syncState = SyncStateExtension.rawValue(item.syncState),
        fileState = FileStateExtension.rawValue(item.fileState),
        syncHasPriority = item.syncHasPriority;

  ItemRecord.fromSyncDict(Map<String, dynamic> dict)
      : rowId = dict['rowId'],
        uid = dict['id'],
        type = dict['type'],
        dateCreated = DateTime.fromMillisecondsSinceEpoch(dict['dateCreated']),
        dateModified =
            DateTime.fromMillisecondsSinceEpoch(dict['dateModified']),
        dateServerModified =
            DateTime.fromMillisecondsSinceEpoch(dict['dateServerModified']),
        deleted = dict['deleted'],
        syncState = SyncState.noChanges,
        fileState = dict['fileState'] ?? FileState.noChanges,
        syncHasPriority = false;

  ItemsCompanion toCompanion() {
    return ItemsCompanion(
        rowId: rowId == null ? const Value.absent() : Value(rowId!),
        id: Value(uid),
        type: Value(type),
        dateCreated: Value(dateCreated),
        dateModified: Value(dateModified),
        dateServerModified: dateServerModified == null
            ? const Value.absent()
            : Value(dateServerModified),
        deleted: Value(deleted),
        syncState: Value(syncState.inString),
        fileState: Value(fileState.inString),
        syncHasPriority: Value(syncHasPriority));
  }

  Future<ItemPropertyRecord?> property(String name,
      [DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    var properties = await db.databasePool.itemPropertyRecordsCustomSelect(
        'name = ? AND item = ?', [Variable(name), Variable(rowId)]);
    if (properties.isNotEmpty) {
      SchemaValueType? valueType =
          GetIt.I<Schema>().expectedPropertyType(type, name);
      if (valueType == null) {
        throw Exception('Not found property $name for $type');
      }
      return ItemPropertyRecord(
          itemRowID: properties[0].item,
          name: properties[0].name,
          value: PropertyDatabaseValue.create(properties[0].value, valueType));
    }
    return null;
  }

  Future<List<ItemPropertyRecord>> properties([DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    var properties = await db.databasePool
        .itemPropertyRecordsCustomSelect('item = ?', [Variable(rowId)]);
    if (properties.isNotEmpty) {
      return properties.map((property) {
        SchemaValueType? valueType =
            GetIt.I<Schema>().expectedPropertyType(type, property.name);
        if (valueType == null) {
          throw Exception(
              'Not found property type ${property.name} for item $type');
        }
        return ItemPropertyRecord(
            itemRowID: property.item,
            name: property.name,
            value: PropertyDatabaseValue.create(property.value, valueType));
      }).toList();
    }
    return [];
  }

  static Future<List<ItemPropertyRecord>> getPropertiesForItems(
      List<ItemRecord> items,
      [DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    var rowIds = items.map((item) => item.rowId);
    var groupedItems = items.toMapByKey((item) => item.rowId);
    var properties = await db.databasePool
        .itemPropertyRecordsCustomSelect("item IN (${rowIds.join(", ")})");
    if (properties.isNotEmpty) {
      return properties.map((property) {
        var type = groupedItems[property.item]!.type;
        SchemaValueType? valueType =
            GetIt.I<Schema>().expectedPropertyType(type, property.name);
        if (valueType == null) {
          throw Exception(
              'Not found property type ${property.name} for item $type');
        }
        return ItemPropertyRecord(
            itemRowID: property.item,
            name: property.name,
            value: PropertyDatabaseValue.create(property.value, valueType));
      }).toList();
    }
    return [];
  }

  Future<PropertyDatabaseValue?> propertyValue(String name,
      [DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    var itemPropertyRecord = await property(name, db);
    if (itemPropertyRecord == null) {
      return null;
    }
    return itemPropertyRecord.value(type, GetIt.I<Schema>());
  }

  Future<int> save([Database? db]) async {
    db ??= AppController.shared.databaseController.databasePool;
    var savedRowID = await db.itemRecordSave(this);
    rowId ??= savedRowID;
    return rowId!;
  }

  setPropertyValue(String name, PropertyDatabaseValue? value,
      {SyncState state = SyncState.update,
      DatabaseController? db,
      bool? isNew}) async {
    db ??= AppController.shared.databaseController;

    /// Create or update the property
    ItemPropertyRecord? itemPropertyRecord;
    if (isNew == null) {
      itemPropertyRecord = await property(name, db);
      isNew = itemPropertyRecord == null;
    }

    itemPropertyRecord ??= ItemPropertyRecord(itemRowID: rowId!, name: name);

    if (value != null) {
      itemPropertyRecord.$value = value;
      await itemPropertyRecord.save(db.databasePool, isNew: isNew);
    } else if (!isNew) {
      await itemPropertyRecord.delete(db.databasePool);
    }

    await saveChanges(state, db);
  }

  setPropertyValueList(List<ItemPropertyRecord> records,
      {SyncState state = SyncState.update, DatabaseController? db}) async {
    db ??= AppController.shared.databaseController;
    for (var record in records) {
      record.itemRowID = rowId;
    }

    await ItemPropertyRecord.insertList(records, db: db.databasePool);

    await saveChanges(state, db);
  }

  Future<int> saveChanges(SyncState state, DatabaseController db) async {
    if (syncState != SyncState.skip) {
      if (syncState == SyncState.create) {
        var item = await ItemRecord.fetchWithUID(uid, db);
        syncState = item!.syncState;
      }
      if (syncState != SyncState.create) {
        syncState = state;
      }
    }

    dateModified = DateTime.now();

    /// Save the item record including the above changes - do this before editing the property so we know the item definitely exists
    return await save(db.databasePool);
    //await addChangeLog(name, value, db); //TODO
  }

  static Future<ItemRecord?> fetchWithUID(String uid,
      [DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    try {
      Item? item = await db.databasePool.itemRecordFetchWithUID(uid);
      if (item == null) return null;
      return ItemRecord.fromItem(item);
    } catch (e) {
      AppLogger.err(e);
      return null;
    }
  }

  static Future<ItemRecord?> fetchWithRowID(int id,
      [DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    try {
      Item? item = await db.databasePool.itemRecordFetchWithRowId(id);
      if (item == null) return null;
      return ItemRecord.fromItem(item);
    } catch (e) {
      AppLogger.err(e);
      return null;
    }
  }

  static Future<List<ItemRecord>> fetchWithUIDs(List<String> uids,
      [Database? db]) async {
    db ??= AppController.shared.databaseController.databasePool;
    try {
      var groupedUids = uids.partition(1000);
      return (await Future.wait(groupedUids
              .map((uids) async => await db!.itemRecordFetchWithUIDs(uids))))
          .expand((element) => element)
          .map((item) => ItemRecord.fromItem(item))
          .toList();
    } catch (e) {
      AppLogger.err(e);
      return [];
    }
  }

  static Future<List<ItemRecord>> fetchWithRowIDs(List<int> ids,
      [Database? db]) async {
    db ??= AppController.shared.databaseController.databasePool;
    try {
      List<Item> items = await db.itemRecordFetchWithRowIDs(ids);
      return items.map((item) => ItemRecord.fromItem(item)).toList();
    } catch (e) {
      AppLogger.err(e);
      return [];
    }
  }

  static Future<List<ItemRecord>> fetchWithType(String type,
      [Database? db]) async {
    db ??= AppController.shared.databaseController.databasePool;
    List<Item> item = await db.itemRecordsFetchByType(type);
    return item.map((e) => ItemRecord.fromItem(e)).toList();
  }

  Future<int> insert(Database db) async {
    return await db.itemRecordInsert(this);
  }

  static Future insertList(List<ItemRecord> records, {Database? db}) async {
    db ??= AppController.shared.databaseController.databasePool;
    var partItems = records.partition(5000);
    for (var part in partItems) {
      await db.itemRecordInsertAll(part);
    }
  }

  Future<int> delete(DatabaseController db) async {
    deleted = true;
    var targetEdges = await edges(null, db: db, deleted: null);
    var sourceEdges = await reverseEdges(null, db: db, deleted: null);
    List<ItemEdgeRecord> relatedEdges = targetEdges + sourceEdges;

    await Future.forEach(
        relatedEdges, (ItemEdgeRecord edge) async => await edge.delete(db));

    return await saveChanges(SyncState.update, db);
  }

  Future<List<ItemEdgeRecord>> edges(String? name,
      {DatabaseController? db, bool? deleted = false}) async {
    db ??= AppController.shared.databaseController;
    try {
      Map<String, dynamic> properties = {'source': rowId};
      if (name != null) {
        properties['name'] = name;
      }
      var edges = await db.databasePool.edgeRecordsSelect(properties);
      var edgeRecords = edges
          .map<ItemEdgeRecord>((edge) => ItemEdgeRecord.fromEdge(edge))
          .toList();
      if (deleted != null) {
        var itemEdgeRecords = (await ItemEdgeRecord.selfItems(edgeRecords, db))
            .toList()
            .toMapByKey((element) => element.rowId);
        edgeRecords = edgeRecords
            .where((edgeRecord) =>
                itemEdgeRecords[edgeRecord.selfRowID]!.deleted == deleted)
            .toList();
      }

      return edgeRecords;
    } catch (e) {
      AppLogger.err(e);
      return [];
    }
  }

  Future<ItemRecord?> edgeItem(String? name,
      {DatabaseController? db,
      bool? deleted = false,
      List<Map<String, dynamic>>? sort}) async {
    var edgeItemRecords =
        await edgeItems(name, db: db, deleted: deleted, limit: 1, sort: sort);
    return edgeItemRecords.asMap()[0];
  }

  Future<ItemEdgeRecord> addEdge(
      {required String edgeName, required ItemRecord targetItem}) async {
    var edgeRecord = ItemEdgeRecord(
        sourceRowID: rowId!,
        name: edgeName,
        targetRowID: targetItem.rowId,
        targetUID: targetItem.uid);
    await edgeRecord.save();
    return edgeRecord;
  }

  Future<List<ItemRecord>> edgeItems(String? name,
      {DatabaseController? db,
      bool? deleted = false,
      int? limit,
      List<Map<String, dynamic>>? sort}) async {
    db ??= AppController.shared.databaseController;
    Map<String, dynamic> properties = {'source': rowId};
    if (name != null) {
      properties['name'] = name;
    }
    return (await db.databasePool.edgeRecordsItemsCustomSelect(properties,
            limit: limit, sort: sort, deleted: deleted))
        .map((targetItem) => ItemRecord.fromItem(targetItem))
        .toList();
  }

  Future<List<ItemEdgeRecord>> reverseEdges(String? name,
      {DatabaseController? db, bool? deleted = false}) async {
    db ??= AppController.shared.databaseController;
    try {
      Map<String, dynamic> properties = {'target': rowId};
      if (name != null) {
        properties['name'] = name;
      }
      var edges = await db.databasePool.edgeRecordsSelect(properties);
      var edgeRecords = edges
          .map<ItemEdgeRecord>((edge) => ItemEdgeRecord.fromEdge(edge))
          .toList();
      if (deleted != null) {
        var itemEdgeRecords = (await ItemEdgeRecord.selfItems(edgeRecords, db))
            .toList()
            .toMapByKey((element) => element.rowId);
        edgeRecords = edgeRecords
            .where((edgeRecord) =>
                itemEdgeRecords[edgeRecord.selfRowID]!.deleted == deleted)
            .toList();
      }

      return edgeRecords;
    } catch (e) {
      AppLogger.err(e);
      return [];
    }
  }

  Future<ItemRecord?> reverseEdgeItem(String? name,
      {DatabaseController? db,
      bool? deleted = false,
      List<Map<String, dynamic>>? sort,
      String? sourceItemType}) async {
    var edgeItemRecords = await reverseEdgeItems(name,
        db: db,
        deleted: deleted,
        limit: 1,
        sort: sort,
        sourceItemType: sourceItemType);
    return edgeItemRecords.asMap()[0];
  }

  Future<List<ItemRecord>> reverseEdgeItems(String? name,
      {DatabaseController? db,
      bool? deleted = false,
      int? limit,
      List<Map<String, dynamic>>? sort,
      String? sourceItemType}) async {
    db ??= AppController.shared.databaseController;
    Map<String, dynamic> properties = {'target': rowId};
    if (name != null) {
      properties['name'] = name;
    }
    return (await db.databasePool.edgeRecordsItemsCustomSelect(properties,
            isReverse: true,
            limit: limit,
            sort: sort,
            deleted: deleted,
            itemType: sourceItemType))
        .map((owningItem) => ItemRecord.fromItem(owningItem))
        .toList();
  }

  factory ItemRecord.fromJson(Map<String, dynamic> json) =>
      _$ItemRecordFromJson(json);

  Map<String, dynamic> toJson() => _$ItemRecordToJson(this);

  @override
  List<Object> get props => [uid, type];

  static Future<List<ItemRecord>> search(
      DatabaseController dbController, String pattern) async {
    List<dynamic> list = await dbController.databasePool
        .itemPropertyRecordsCustomSelect(
            'value MATCH ?', [Variable.withString(pattern)], true);
    return (await Future.wait(list.map((el) async {
      Item? item = await dbController.databasePool
          .itemRecordFetchWithRowId(int.parse(el.item));
      if (item == null) return null;
      return ItemRecord.fromItem(item);
    })))
        .whereType<ItemRecord>()
        .toList();
  }

  static Future<ItemRecord?> me([DatabaseController? dbController]) async {
    dbController ??= AppController.shared.databaseController;
    var edge = await dbController.databasePool.edgeRecordSelect({'name': 'me'});
    if (edge == null) {
      return null;
    }

    return await ItemRecord.fetchWithRowID(edge.target, dbController);
  }

  static Future<ItemRecord?> createMe(
      [DatabaseController? dbController]) async {
    dbController ??= AppController.shared.databaseController;
    var myself = ItemRecord(type: 'Person');
    await myself.save(dbController.databasePool);

    await ItemEdgeRecord(
            sourceRowID: myself.rowId, name: 'me', targetRowID: myself.rowId)
        .save(dbController.databasePool);
    return myself;
  }

  addChangeLog(String name, PropertyDatabaseValue? value,
      [DatabaseController? dbController]) async {
    dbController ??= AppController.shared.databaseController;
    var auditItem = ItemRecord(type: 'AuditItem');
    await auditItem.save(dbController.databasePool);
    if (rowId == null || auditItem.rowId == null) {
      throw Exception(
          "Add changelog: Item doesn't have row id, possibly not saved?");
    }

    await auditItem.setPropertyValue(
        'date', PropertyDatabaseValueDatetime(DateTime.now()),
        db: dbController);
    await auditItem.setPropertyValue(
        'content', PropertyDatabaseValueString(name),
        db: dbController);
    await auditItem.setPropertyValue(
        'action', PropertyDatabaseValueString('edit'),
        db: dbController);
    await ItemEdgeRecord(
            sourceRowID: rowId, name: 'changelog', targetRowID: auditItem.rowId)
        .save(dbController.databasePool);
  }

  static final Map<String, String> _mapSchemaValueType = {
    'string': 'Text',
    'int': 'Integer',
    'double': 'Real',
    'bool': 'Bool',
    'datetime': 'DateTime',
  };

  static String? mapSchemaValueType(String nativeType) {
    return _mapSchemaValueType[nativeType];
  }

  static String? reverseMapSchemaValueType(String propertyValue) {
    return _mapSchemaValueType.keys.firstWhere(
        (nativeType) => _mapSchemaValueType[nativeType] == propertyValue,
        orElse: () => propertyValue);
  }

  static final Map<String, String> _mapSchemaPropertyName = {
    'action': 'actionName',
    'query': 'queryName',
    'key': 'keyName',
    'type': 'typeName',
  };

  static String mapSchemaPropertyName(String propertyName) {
    return _mapSchemaPropertyName[propertyName] ?? propertyName;
  }

  static String reverseMapSchemaPropertyName(String propertyName) {
    return _mapSchemaPropertyName.keys.firstWhere(
        (nativePropertyName) =>
            _mapSchemaPropertyName[nativePropertyName]?.toLowerCase() ==
            propertyName.toLowerCase(),
        orElse: () => propertyName);
  }

  static getPropertyFromSync(itemType, propertyName, propertyValue) {
    switch (itemType) {
      case 'ItemPropertySchema':
        switch (propertyName) {
          case 'valueType':
            propertyValue = reverseMapSchemaValueType(propertyValue);
            break;
          case 'propertyName':
            propertyValue = reverseMapSchemaPropertyName(propertyValue);
            break;
        }
        break;
      case 'ItemEdgeSchema':
        switch (propertyName) {
          case 'edgeName':
            propertyValue = reverseMapSchemaPropertyName(propertyValue);
            break;
        }
        break;
    }

    return propertyValue;
  }

  Map<String, dynamic> mergeDict(
      {required List<ItemPropertyRecord> properties, required Schema schema}) {
    Map<String, dynamic> keyProperties = {
      'type': type,
      'id': uid,
      'dateCreated': dateCreated.millisecondsSinceEpoch,
      'dateModified': dateModified.millisecondsSinceEpoch,
      'dateServerModified': dateServerModified?.millisecondsSinceEpoch,
      'deleted': deleted,
    };

    for (var element in properties) {
      var propertyValue = element.$value.value;
      keyProperties.addEntries([MapEntry(element.name, propertyValue)]);
    }

    return keyProperties;
  }

  Future<Map<String, dynamic>> syncDict(DatabaseController dbController) async {
    return mergeDict(
        properties: await properties(dbController), schema: GetIt.I<Schema>());
  }

  static deleteExistingDBKeys([DatabaseController? db]) async {
    db ??= AppController.shared.databaseController;
    var dbKeys = await db.databasePool.itemRecordsFetchByType('CryptoKey');
    if (dbKeys.isNotEmpty) {
      await Future.forEach(dbKeys,
          (Item item) async => await ItemRecord.fromItem(item).delete(db!));
    }
  }

  //TODO: return private keys
  static setOwnerAndDBKey(
      {required String privateKey,
      required String publicKey,
      required String dbKey,
      Database? db}) async {
    db ??= AppController.shared.databaseController.databasePool;
    try {
      await ItemRecord.deleteExistingDBKeys();
      var meRowId = (await ItemRecord.me())?.rowId ??
          (await ItemRecord.createMe())?.rowId;
      if (meRowId == null) {
        throw Exception('Could not find me user');
      }

      await AuthKey(
              ownerId: meRowId,
              type: '64CharacterRandomHex',
              role: null,
              key: dbKey,
              name: 'Memri Database Key',
              active: true)
          .save(db);
      /*var privateKeyItem = await AuthKey(
              ownerId: meRowId,
              type: "ED25519",
              role: "private",
              key: privateKey,
              name: "Memri Owner Key",
              active: true)
          .save(db);*/
      var publicKeyItem = await AuthKey(
              ownerId: meRowId,
              type: 'ED25519',
              role: 'public',
              key: publicKey,
              name: 'Memri Owner Key',
              active: true)
          .save(db);

      /*await ItemEdgeRecord(sourceRowID: privateKeyItem!.rowId, name: "owner", targetRowID: meRowId)
          .save(db);*/
      await ItemEdgeRecord(
              sourceRowID: publicKeyItem!.rowId,
              name: 'owner',
              targetRowID: meRowId)
          .save(db);
      /*await ItemEdgeRecord(
              sourceRowID: privateKeyItem.rowId,
              name: "publicKey",
              targetRowID: publicKeyItem.rowId)
          .save(db);
      await ItemEdgeRecord(
              sourceRowID: publicKeyItem.rowId,
              name: "privateKey",
              targetRowID: privateKeyItem.rowId)
          .save(db);*/
    } catch (error) {
      AppLogger.err('ERROR: setOwnerAndDBKey $error');
      throw Exception('Error deleting existing db keys');
    }
  }

  static Future<AuthKeys> getOwnerAndDBKey([Database? db]) async {
    db ??= AppController.shared.databaseController.databasePool;
    var meRowId = (await ItemRecord.me())?.rowId;
    if (meRowId == null) {
      throw Exception('Could not find me user');
    }

    var items = await db.itemRecordsFetchByType('CryptoKey');
    ItemRecord? ownerKey, dbKey;
    String? ownerKeyValue, dbKeyValue;
    if (items.isNotEmpty) {
      for (var item in items) {
        var itemRecord = ItemRecord.fromItem(item);
        if (await itemRecord.propertyValue('role') ==
            PropertyDatabaseValueString('public')) {
          if (await itemRecord.propertyValue('active') ==
              PropertyDatabaseValueBool(true)) {
            if ((await itemRecord.edgeItem('owner'))?.rowId == meRowId) {
              ownerKey = itemRecord;
              ownerKeyValue = (await ownerKey.propertyValue('keystr'))?.value;
            }
          }
        }
        if (await itemRecord.propertyValue('name') ==
            PropertyDatabaseValueString('Memri Database Key')) {
          if (await itemRecord.propertyValue('active') ==
              PropertyDatabaseValueBool(true)) {
            dbKey = itemRecord;
            dbKeyValue = (await dbKey.propertyValue('keystr'))?.value;
          }
        }
      }
    }
    if (ownerKey == null ||
        dbKey == null ||
        ownerKeyValue == null ||
        dbKeyValue == null) {
      throw Exception('Keys not found');
    }
    return AuthKeys(ownerKey: ownerKeyValue, dbKey: dbKeyValue);
  }

  //TODO: copy also edges (would be recursive)
  Future<ItemRecord> copy(DatabaseController db,
      {Map<String, PropertyDatabaseValue>? withProperties}) async {
    var newItem = ItemRecord(type: type);
    await newItem.save();
    var props = await properties(db);
    for (var element in props) {
      element.itemRowID = newItem.rowId!;
    }
    withProperties?.forEach((propName, propValue) {
      props.add(ItemPropertyRecord(name: propName, value: propValue));
    });
    await newItem.setPropertyValueList(props);
    return newItem;
  }
}
