import 'package:memri/core/apis/pod_api.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/pod/pod_config.dart';
import 'package:memri/core/services/api_service.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/network_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:collection/collection.dart';

class GraphService extends ApiService<PodAPI> {
  GraphService(this._prefs) : super(api: PodAPI()) {
    _setConfig();
  }

  final SharedPreferences _prefs;
  late PodConfig podConfig;

  Future<List<Item>> getResultAsItemList({required String query}) async {
    _setConfig();
    return NetworkHelper().parseGQLResponse(await api.queryGraphQL(query));
  }

  void _setConfig() {
    podConfig = PodConfig(
      baseUrl:
          _prefs.getString(app.keys.podAddress) ?? app.settings.defaultPodUrl,
      ownerKey: _prefs.getString(app.keys.ownerKey) ?? '',
      databaseKey: _prefs.getString(app.keys.dbKey) ?? '',
      apiVersion: app.settings.podVersion,
    );
    api.setConnectionConfig(podConfig);
  }

  Future<List<Item>> fetchActiveImporters() async =>
      await getResultAsItemList(query: '''
query {
        PluginRun {
          id
          pluginName
          status
        }
      }
''');

  Future<List<Item>> checkIfDataInPod(String source) async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "$source"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchWhatsappMessages() async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "whatsapp"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchSampleWhatsappMessages() async =>
      await getResultAsItemList(query: '''
query {
        Message(limit: 1, filter: {service: {eq: "sample-whatsapp"}}) {
          id
          service
        }
      }
''');

  Future<List<Item>> fetchTwitterTweets() async =>
      await getResultAsItemList(query: '''
query {
        Tweet(limit: 1) {
          id
        }
      }
''');

  Future<List<Item>> getOAuthFlowItems(String service) async =>
      await getResultAsItemList(query: '''
      query {
        OauthFlow (limit: 500 filter: {service: {eq: "$service"}}) {
          accessToken
          expiresIn
          createdAt
          refreshToken
          dateCreated
          service
        }
      }''');

  Future<List<Item>> getProjectItems(String projectId) async =>
      // TODO split with `..on Message and ..on Tweet`
      await getResultAsItemList(query: '''
      query {
        Project (filter: {id: {eq: "$projectId"}}) {
          name
          dataSource
          dataset {
            name
            entry {
              skippedByLabeller
              data {
                content
                message
                dateSent
                postDate
                author {
                  displayName
                  handle
                }
              }
              annotation {
                labelValue
              }
            }

            labellingTask {
              id
              labelOption {
                name
              }
            }
          }
        }
      }
    ''');

  Future<List<Item>> getAllProjectsNames() async =>
      await getResultAsItemList(query: '''
      query {
        Project (filter: {deleted: {eq: 0}}){
          name
        }
      }''');

  Future<List<Item>> getProjectSummaries() async =>
      await getResultAsItemList(query: '''
      query {
        Project(limit: 1000, filter: {deleted: {eq: 0}}) {
          id
          name
          dataset (limit: 1) {
            name
            entry {
              skippedByLabeller
              annotation {
                labelValue
              }
            }
          }
        }
      }''');

  Future<List<Item>> getMessages(String service) async =>
      await getResultAsItemList(query: '''
      query {
        Message (limit: 500, filter: {service: {eq: "$service"}}) {
          id
          content
          dateSent
          service
        }
      }''');

  Future<List<Item>> getTweets() async => await getResultAsItemList(query: '''
      query {
        Tweet (limit: 500) {
          id
          message
          postDate
        }
      }''');

  Future<List<Item>> getMockItems(
      String sourceProject, String type, List<String> field) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        $type(order_desc: dateSent, filter: {sourceProject: {eq: "$sourceProject"}}) {
          ${field.join('\n')}
          dateSent
          sourceProject
        }
      }
    ''');
    result = result.where((x) => x.get('deleted') != true).toList();

    return result;
  }

  Future<List<Item>> LoadProjectItems(
      String memriProjectId, int amount, List<String> properties) async {
    String sorter = '';
    if (properties.contains('dateSent')) {
      sorter = 'order_desc: dateSent,';
    } else if (properties.contains('postDate')) {
      sorter = 'order_desc: postDate,';
    }
    List<Item> res = await getResultAsItemList(query: '''
      query {
        Project(filter: {id: {eq: "$memriProjectId"}}) {
          dataset{
            id
            entry {
              data($sorter, filter: {deleted: {eq: 0}}) {
                ${properties.join('\n')}
                isMock
              }
            }
          }
        }
      }
    ''');
    List<Item> entries = res[0].edgeItem('dataset')!.getEdgeTargets('entry')!;
    List<Item> resItems = entries.map((e) => e.edgeItem('data')!).toList();
    //ugly hack to make sure we can filter and limit
    resItems = resItems.where((x) => x.get('isMock') != true).toList();
    if (resItems.length > amount) {
      resItems = resItems.sublist(0, amount);
    }
    return resItems;
  }

  Future<List<Item>> loadAppContent(
      String pluginId, String dataType, List<String> properties) async {
    String sorter = '';
    if (properties.contains('dateSent')) {
      sorter = 'order_desc: dateSent,';
    } else if (properties.contains('postDate')) {
      sorter = 'order_desc: postDate,';
    }
    String edges = '';
    if (dataType == 'Tweet') {
      edges = '''
        author {
          displayName
          handle

        }
      ''';
    }
    List<Item> result = await getResultAsItemList(query: '''
      query {
        $dataType($sorter filter: {and: [
          {sourceProject: {eq: "$pluginId"}},
          {deleted: {eq: 0}}
        ]}   
        ) {
          ${properties.join('\n')}
          sourceProject
          label {
            value
          }
          $edges
        }
      }
    ''');
    // hack untill we have boolean filters in graphql
    result = result.where((x) => x.get('deleted') != true).toList();

    return result;
  }

  Future<List<Item>> fetchDataApps() async =>
      await getResultAsItemList(query: '''
      query {
        Plugin {
          containerImage
          configJson
          config
          dataType
          pluginDescription
          name
          pluginModule
          pluginName
          pluginType
          gitProjectId
        }
      }
    ''');

  Future<List<Item>> getInboxChats() async =>
      await getResultAsItemList(query: '''
      query {
        MessageChannel (limit: 1000, filter: {service: {eq: "whatsapp"}}) {
          id
          name
          service
          photo {
            id
            file {
              sha256
            }
          }
          ~messageChannel (limit: 1, order_desc: dateSent) {
            dateSent
          }
        }
      }''');

  Future<List<Item>> getInboxMessages(
          String chatID, int limit, int offset) async =>
      await getResultAsItemList(query: '''
      query {
        MessageChannel (filter: {id: {eq: "$chatID"}}) {
          id
          ~messageChannel (
            limit: $limit,
            offset: $offset,
            order_desc: dateSent,
            filter: {or: [{isMock: {eq: null}}, {isMock: {eq: 0}}]}
          ) {
            dateSent
            content
            isMock
            sender {
              displayName
                isMe
                profilePicture {
                id
                file {
                  sha256
                }
              }
            }
          }
        }
      }
    ''');

  Future<List<Item>> getTwitterFeed() async =>
      await getResultAsItemList(query: '''
      query {
        Tweet (
          limit: 1000,
          order_desc: postDate,
          filter: {or: [{isMock: {eq: null}}, {isMock: {eq: 0}}]}
        ) {
          externalId
          message
          isMock
          author {
            handle
            displayName
            profilePicture {
                id
                file {
                  sha256
                }
            }
          }
          mention {
            handle
            displayName
          }
          postDate
        }
      }''');

  Future<Item?> getPropertySchema(String itemType, String propertyName) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        ItemPropertySchema(limit: 1, filter: {and: [
              {itemType: {eq: "$itemType"}},
              {propertyName: {eq: "$propertyName"}}
            ]}) {
          itemType
          propertyName
        }
      }
      ''');
    return result.firstOrNull;
  }

  Future<Item?> getPluginRunsWithContainerImage(String containerImage) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        PluginRun(filter: {containerImage: {eq: "$containerImage"}}, order_desc: dateCreated) {
              id
              containerImage
              config
              containerId
              status
        }
      }
      ''');
    return result.firstOrNull;
  }

  Future<Item?> getProjectFromPlugin({required id}) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        Plugin(filter: {id: {eq: "$id"}}) {
              id
              project {
                id
                dataSource
              }
        }
      }
      ''');
    return result.firstOrNull?.edgeItem('project');
  }

  Future<List<Item>> getPluginRunsForPlugin(String pluginId) async {
    List<Item> result = await getResultAsItemList(query: '''
      query {
        Plugin(filter: {id: {eq: "$pluginId"}}, order_desc: dateCreated) {
              run {
                id
              }
        }
      }
      ''');
    return result.firstOrNull!.getEdgeTargets('run')!;
  }
}
