import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/services.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/apis/pod/pod_payloads.dart';
import 'package:memri/core/apis/pod_api.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/pod/pod_config.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/core/services/api_service.dart';
import 'package:memri/utilities/extensions/string.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/network_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:uuid/uuid.dart';
import 'package:uuid/uuid_util.dart';

class PodService extends ApiService<PodAPI> {
  PodService(this._prefs) : super(api: PodAPI()) {
    podConfig = PodConfig(
      baseUrl:
          _prefs.getString(app.keys.podAddress) ?? app.settings.defaultPodUrl,
      ownerKey: _prefs.getString(app.keys.ownerKey) ?? '',
      databaseKey: _prefs.getString(app.keys.dbKey) ?? '',
      apiVersion: app.settings.podVersion,
    );
    api.setConnectionConfig(podConfig);
  }

  final SharedPreferences _prefs;
  late PodConfig podConfig;

  ///  Authentication
  Future<void> authenticate({
    required String podAddress,
    required String ownerKey,
    required String dbKey,
  }) async {
    podConfig = PodConfig(
      baseUrl: podAddress,
      ownerKey: ownerKey,
      databaseKey: dbKey,
      apiVersion: app.settings.podVersion,
    );
    api.setConnectionConfig(podConfig);
    await api.authenticate();
    await storePodAddress(podAddress);
    await storeOwnerKey(ownerKey);
    await storeDatabaseKey(dbKey);
  }

  Future<void> storeOwnerKey(String ownerKey) async =>
      await _prefs.setString(app.keys.ownerKey, ownerKey);

  Future<void> storeDatabaseKey(String dbKey) async =>
      await _prefs.setString(app.keys.dbKey, dbKey);

  Future<void> storePodAddress(String podAddress) async =>
      await _prefs.setString(app.keys.podAddress, podAddress);

  String generateCryptoStrongKey() {
    return "${Uuid().v4(options: {
          'rng': UuidUtil.cryptoRNG
        })}${Uuid().v4(options: {'rng': UuidUtil.cryptoRNG})}"
        .replaceAll('-', '');
  }

  /// POD

  Future<void> createSchema() async {
    List<Map<String, dynamic>> schemaItems = List.from(
        jsonDecode(await rootBundle.loadString('assets/schema.json')));

    AppLogger.debug('create schema: ${schemaItems.length} items');
    var payload = PodPayloadBulkAction(
        createItems: schemaItems,
        updateItems: [],
        deleteItems: [],
        createEdges: []);
    api.bulkAction(payload);
  }

  Future<String> podVersion() async =>
      api.podVersion().catchError((error) => '');

  Future<void> deleteItem(String id) => bulkAction(deleteItems: [id]);

  Future<dynamic> bulkAction({
    List<Item>? createItems,
    List<Item>? updateItems,
    List<String>? deleteItems,
    List<Edge>? createEdges,
  }) async {
    List<Map<String, dynamic>> createPayload = [];
    if (createItems != null) {
      for (var item in createItems) {
        item.setIdIfNotExists();
        createPayload.add(item.toJson());
      }
    }

    List<Map<String, dynamic>> updatePayload = [];
    if (updateItems != null) {
      for (var item in updateItems) {
        updatePayload.add(item.toJson());
      }
    }

    List<Map<String, dynamic>> edgePayload = [];
    if (createEdges != null) {
      for (var edge in createEdges) {
        edgePayload.add(edge.toJson());
      }
    }

    var bulkPayload = PodPayloadBulkAction(
        createItems: createPayload,
        updateItems: updatePayload,
        deleteItems: deleteItems ?? [],
        createEdges: edgePayload);

    return await api.bulkAction(bulkPayload);
  }

  Future<List<Item>> search(dynamic paylod) async {
    final response = await api.search(paylod);
    final items = List<Item>.from(response.map((item) => Item.fromJson(item)));
    return items;
  }

  Future<Item?> getItem({
    required String id,
  }) async {
    try {
      return Item.fromJson(await api.getItem(id));
    } catch (e) {
      return null;
    }
  }

  /// TODO @Deprecated('This method moved to the GraphService')
  Future<List<Item>> graphql({required String query}) async =>
      NetworkHelper().parseGQLResponse(await api.queryGraphQL(query));

  Future<Item> createItem({
    required Item item,
  }) async {
    // if (item.type == null) {
    //   throw new Exception("Attempted to create item without an item type.");
    // }
    var itemMap = item.toJson();
    var resultID = await api.createItem(itemMap);
    item.properties['id'] = resultID;
    return item;
  }

  Future<Item> updateItem({
    required Item item,
  }) async {
    var itemMap = item.toJson();
    await api.updateItem(itemMap);
    return item;
  }

  Future<void> removeEdge(Edge edge) async {
    var sourceId = edge.source.get('id');
    var targetId = edge.target.get('id');
    // TODO add to PodPayloads
    Map<String, String> payload = {
      '_source': sourceId,
      '_target': targetId,
    };

    return api.deleteEdge(payload);
  }

  Future<String> getLogsForPluginRun(String id) async {
    var logsJson = await api.getLogsForPluginRun(id);
    return logsJson['logs'];
  }

  Future<Uint8List> downloadFile(String fileSHAHash) async =>
      await api.downloadFile(fileSHAHash);

  Future<Item> createDataset(Item currentProject) async {
    var dataset = Item(type: 'Dataset');
    var name =
        '${(datasetAdjectives..shuffle()).first.capitalize()} ${(datasetNouns..shuffle()).first.capitalize()}';
    dataset.properties['name'] = name;
    var datasetEdge =
        Edge(source: currentProject, target: dataset, name: 'dataset');
    await bulkAction(createItems: [dataset], createEdges: [datasetEdge]);
    return dataset;
  }

  Future<Item> startImporter(String pluginName, String pluginModule,
      String container, String mixPanelName) async {
    String id = app.randomUuid();

    Item item = Item(
      type: 'PluginRun',
      properties: {
        'id': id,
        'pluginName': pluginName,
        'pluginModule': pluginModule,
        'containerImage': container,
        'status': 'idle',
        'targetItemId': id
      },
    );
    await bulkAction(createItems: [item]);
    AnalyticsService().logImporterConnect(mixPanelName);
    return item;
  }

  Future<String?> oauth2Authorize(
      String code, String redirectUrl, String platform) async {
    dynamic resMap = await api.oauth2Authorize(code, redirectUrl, platform);
    return resMap['accessToken'];
  }

  Future<String?> oauth2GetAccessToken(String platform) async {
    dynamic resMap = await api.oauth2GetAccessToken(platform);
    return resMap['accessToken'];
  }

  Future<String> oauth2url(
      String redirectUrl, List<String> scopes, String service) async {
    return await api.oauth2url(redirectUrl, scopes, service);
  }
}

List<String> datasetAdjectives = [
  'precious',
  'supreme',
  'succinct',
  'tiresome',
  'plausible',
  'exclusive',
  'screeching',
  'nauseating',
  'worthless',
  'endurable',
  'powerful',
  'shocking',
  'shallow',
  'industrious',
  'weary',
  'tacky',
  'drunk',
  'boorish',
  'humorous',
  'lame'
];
List<String> datasetNouns = [
  'dolphin',
  'monkey',
  'lizard',
  'chi oua oua',
  'butterfly',
  'camel',
  'goldfish',
  'piranha',
  'elephant',
  'giraffe',
  'baboon',
  'bobcat',
  'dingo',
  'gorilla'
];
