import 'dart:async';
import 'dart:convert';
import 'dart:math';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/models/running_plugin.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/utilities/helpers/time_helper.dart';
import 'package:visibility_detector/visibility_detector.dart';

enum AppState { deploying, initialPredictions, showApp, showLogs, showSettings }

extension ParseToString on AppState {
  String toUrlString() {
    switch (this) {
      case AppState.deploying:
        return 'loading';
      case AppState.initialPredictions:
        return 'predicting';
      case AppState.showApp:
        return 'app';
      case AppState.showLogs:
        return 'logs';
      case AppState.showSettings:
        return 'settings';
    }
  }
}

enum AppListState { fetching, fetched }

class DataAppProvider with ChangeNotifier {
  final PodService _podService;
  final GraphService _graphService;
  final GitlabService _gitlabService;

  DataAppProvider(this._podService, this._graphService, this._gitlabService);

  // TODO handle schema checking somewhere else
  final List<String> _runProperties = [
    'containerImage',
    'authUrl',
    'pluginModule',
    'pluginName',
    'status',
    'targetItemId',
    'error',
    'config',
    'progress',
  ];

  final List<String> _pluginProperties = [
    'bundleImage',
    'containerImage',
    'configJson',
    'config',
    'dataType',
    'icon',
    'pluginDescription',
    'name',
    'pluginModule',
    'pluginName',
    'pluginType',
    'gitProjectId',
  ];

  Map<String, String> source2type = {
    'twitter': 'Tweet',
    'whatsapp': 'Message',
    'sampleWhatsapp': 'Message'
  };

  // Map<String, String> source2type = {'Tweet': 'twitter', 'Message': 'Message', 'sampleWhatsapp': 'Message'};

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? feedStreamSubscription;
  List<Item> feed = [];

  String? currentDataType;
  String? currentDataSource;
  Item? _currentMemriProject;
  Item? currentPlugin;

  set currentMemriProject(Item? item) {
    if (item != null) {
      _currentMemriProject = item;
      currentDataSource = _currentMemriProject?.get('dataSource');
      currentDataType = source2type[currentDataSource]!;
    }
  }

  Item? get currentMemriProject => _currentMemriProject;

  String? get currentMemriProjectId => _currentMemriProject?.get('id');

  String getFeedPostDate(int feedIndex) =>
      DateTimeHelper().getDateString(feed[feedIndex].get('postDate'));

  String getFeedAuthorName(int feedIndex) =>
      feed[feedIndex].edgeItem('author')?.get('displayName') ?? '';

  String getFeedAuthorHandle(int feedIndex) =>
      feed[feedIndex].edgeItem('author')?.get('handle') ?? '';

  String? get mockItemDataSourceForGit =>
      currentGitProjectId == null ? null : '_preview-$currentGitProjectId';

  reset() {
    currentPluginRun = null;
    currentPluginRunStream = null;
    currentAppState = AppState.deploying;
    currentRunLog = null;
    currentGitProjectId = null;
    currentDataAppItems = [];

    //Inbox
    chatStream = null;
    messageStream = null;
    if (chatStreamSubscription != null) chatStreamSubscription!.cancel();
    if (messageStreamSubscription != null) messageStreamSubscription!.cancel();
    if (feedStreamSubscription != null) feedStreamSubscription!.cancel();
    if (currentPluginRunSubscription != null) {
      currentPluginRunSubscription!.cancel();
    }
    selectedChannel = -1;
    chats = [];
    activeChatMessages = [];
    _loadingMessages = false;
    _loadingMoreMessages = false;
    _noChatRoomsFound = false;
    chatImageSha = {};
    chatImageProviders = {};
    messageImageSha = {};
    tweetImageSha = {};
    messageImageProviders = {};
    tweetImageProviders = {};
    selectedChatRoomID = null;
    dataApps = [];
    feed = [];
    currentDataType = null;
    currentDataSource = null;
    _currentMemriProject = null;
    currentPlugin = null;
  }

  /*
   * Inbox
   */

  Future<Uint8List> downloadPhoto(String fileSHAHash) async =>
      await _podService.downloadFile(fileSHAHash);

  /*
  * Running a plugin
  */

  Item? currentPluginRun;
  Stream<Item>? currentPluginRunStream;
  StreamSubscription<Item>? currentPluginRunSubscription;
  AppState currentAppState = AppState.showApp;
  AppListState appListState = AppListState.fetching;
  String? currentRunLog;
  int? currentGitProjectId;
  List<Item> currentDataAppItems = [];

  setDataAppState(AppState state) async {
    // first show the logs are loading
    currentAppState = state;

    /// TODO: FIX THIS, CAUSE EXCEPTION
    notifyListeners();

    // then, show log contents
    if (state == AppState.showLogs) {
      currentRunLog = await getRunLog(currentPluginRun!);
    }
    notifyListeners();
  }

  handlePublishApp(BuildContext context, Item project, String dataSource) {
    // TODO delete previous app with name + gitProjectId

    createAndConnectPluginItem(project);
    RouteNavigator.navigateTo(context: context, route: Routes.apps);
  }

  // handleRerunApp(Item project) {
  //   runIndexerPluginFromGit(project: project, changeState: false);
  // }

  createAndConnectPluginItem(project) async {
    String projectName = project.get('name')!;
    // Create data app from current gitlab id
    var metadata = await getMetadata(currentGitProjectId!);
    var plugin = Item(type: 'Plugin');
    metadata.forEach((key, value) {
      if (_pluginProperties.contains(key)) {
        if (key == 'containerImage' && value is String) {
          value = value.toLowerCase();
        }
        plugin.properties[key] = value;
      }
    });

    plugin.properties['name'] = projectName;
    plugin.properties['pluginDescription'] = null;
    plugin.properties['gitProjectId'] = currentGitProjectId;
    // plugin.properties['dataType'] = 'Text';
    plugin.properties['pluginType'] = 'dataApp';
    plugin.properties['dataType'] = currentDataType;
    plugin.setIdIfNotExists();
    Edge edge = Edge(name: 'project', source: plugin, target: project);

    await createMockItems(128, plugin.id);
    // await _podService.createItem(item: plugin);
    await _podService.bulkAction(createItems: [plugin], createEdges: [edge]);
    dataApps.add(plugin);
    AppLogger.debug('Created plugin: $projectName');
  }

  runIndexerPluginFromGit(
      {required Item project, bool changeState = true}) async {
    currentPluginRun = null;

    String? dataSource = mockItemDataSourceForGit;
    if (dataSource == null) {
      debugPrint('Could not run plugin, git project id not set');
    } else {
      currentDataAppItems = await createMockItems(10, dataSource);

      debugPrint('${currentDataAppItems.length} mock items available');

      var pluginRun =
          await createPluginRunFromGit(currentGitProjectId!, currentDataType!);
      await _podService.createItem(item: pluginRun);
      AppLogger.debug('Running plugin: ${pluginRun.toJson()}');

      currentPluginRun = pluginRun;
      downloadMockItemsWhenPluginFinished(
          runId: currentPluginRun!.get('id'),
          dataType: currentDataType!,
          changeState: changeState);
    }
  }

  Stream<Item> _pluginRunStream(String runId) async* {
    while (true) {
      var run = await _podService.getItem(id: runId);
      if (run != null) {
        yield run;
      }
      await Future.delayed(Duration(seconds: 1));
    }
  }

  bool runIsDone(String? runStatus) => ['done', 'error'].contains(runStatus);

  downloadMockItemsWhenPluginFinished(
      {required String runId,
      required String dataType,
      bool changeState = true}) {
    currentPluginRunStream = _pluginRunStream(runId);
    currentPluginRunSubscription = currentPluginRunStream!.listen((run) async {
      currentPluginRun = run;
      AppLogger.debug('run status: ${getCurrentRunStatus()}');
      // required to set the state of the screen
      currentRunLog = await getRunLog(currentPluginRun!);
      notifyListeners();

      if (runIsDone(getCurrentRunStatus())) {
        stopPluginRunStream();
        // loadCurrentAppContent();
        if (changeState) {
          currentAppState = AppState.showApp;
        }
        currentDataAppItems =
            await loadAppContent('_preview-$currentGitProjectId', dataType);
        for (var item in currentDataAppItems) {
          if (item.edgeItem('label') != null) {
            AppLogger.debug(item.edgeItem('label')!.get('value'));
          }
        }
        notifyListeners();
        AppLogger.debug('Run done');
        AppLogger.debug('app message: ${currentDataAppItems.length}');
        AppLogger.debug('has log: ${currentRunLog != null}');
      } else if (getCurrentRunStatus() == 'started') {
        if (changeState) {
          currentAppState = AppState.initialPredictions;
          notifyListeners();
        }
      }
    });
  }

  stopPluginRunStream() {
    currentPluginRunSubscription?.cancel();
    currentPluginRunStream = null;
  }

  Future<Map<String, dynamic>> getMetadata(int gitProjectId) async {
    // TODO error handling if metadata does not exist
    AppLogger.debug('fetching metadata...');
    var fileContent = await _gitlabService.getTextFileContent(
        projectId: gitProjectId, filepath: 'metadata.json');
    return jsonDecode(fileContent);
  }

  Future<Item> createPluginRunFromGit(int gitProjectId, String itemType) async {
    var metadata = await getMetadata(gitProjectId);

    var pluginRun = Item(type: 'PluginRun');
    metadata.forEach((key, value) {
      if (_runProperties.contains(key)) {
        // containerImage needs to be lowercase
        if (key == 'containerImage' && value is String) {
          value = value.toLowerCase();
        }
        pluginRun.properties[key] = value;
      }
    });

    // Set id separately, pluginrun needs targetItemId == id
    pluginRun.setIdIfNotExists();
    pluginRun.properties['targetItemId'] = pluginRun.get('id');
    pluginRun.properties['status'] = 'not initialised';
    pluginRun.properties['config'] =
        jsonEncode({'isMock': true, 'item_type': itemType});

    return pluginRun;
  }

  Future<String?> getRunLog(Item run) async =>
      await _podService.getLogsForPluginRun(run.get('id'));

  bool pluginIsNotRunning() {
    String? status = getCurrentRunStatus();
    // status is null means no plugin (but maybe an older run exists)
    if (status == 'done' || status == null || status == 'error') {
      return true;
    }
    return false;
  }

  String? getCurrentRunStatus() {
    // TODO run status enum, sync with pymemri.plugin.states
    return currentPluginRun?.get('status');
  }

  Future<List<Item>> createMockItems(int amount, String sourceProject) async {
    List<String> fields;
    if (currentDataType == 'Message') {
      fields = ['content', 'dateSent'];
    } else if (currentDataType == 'Tweet') {
      fields = ['message', 'postDate'];
    } else {
      throw Error();
    }

    var existingMockItems = await _graphService.getMockItems(
        sourceProject, currentDataType!, fields);
    // AppLogger.debug(existingMockItems);
    int amountToCreate = max((amount - existingMockItems.length), 0);

    var items = await _graphService.LoadProjectItems(
        currentMemriProjectId!, amountToCreate, fields);
    List<Item> newItems = [];
    for (var item in items) {
      Map<String, dynamic> fieldValues = {};
      for (var field in fields) {
        var fieldValue = item.get(field);
        if (fieldValue != null) {
          fieldValues[field] = fieldValue;
        }
      }
      Map<String, dynamic> json = {
        'type': currentDataType,
        'isMock': true,
        'sourceProject': sourceProject
      };
      json.addAll(fieldValues);
      var newItem = Item.fromJson(json);
      if (currentDataType == 'Message') {
        newItem.properties['dateSent'] = item.get('dateSent');
      }
      newItems.add(newItem);
    }
    // AppLogger.debug(newItems);

    await _podService.bulkAction(createItems: newItems);
    AppLogger.debug('Created ${newItems.length} new mock messages');
    AppLogger.debug(
        'using ${existingMockItems.length} existsing mock messages');
    return existingMockItems + newItems;
  }

  Future<List<Item>> loadAppContent(String pluginId, String dataType,
      {bool onlyWithPredictions = true}) async {
    List<String> properties;
    if (dataType == 'Message') {
      properties = ['content', 'dateSent'];
    } else if (dataType == 'Tweet') {
      properties = ['message', 'postDate'];
    } else {
      throw Error();
    }
    var response =
        await _graphService.loadAppContent(pluginId, dataType, properties);

    List<Item> result;
    if (onlyWithPredictions) {
      result = [];
      for (var element in response) {
        if (element.edgeItem('label')?.get('value') != null) {
          result.add(element);
        }
      }
    } else {
      result = response;
    }

    AppLogger.debug(
        'loaded ${result.length} items for preview out of ${response.length} total (with or without label)');

    notifyListeners();

    return result;
  }

  /*
  * Apps screen
  */

  // pluginId: runningPlugin
  List<Item> dataApps = [];

  fetchDataApps() async {
    _handleAppListFetching();

    dataApps =
        await _graphService.fetchDataApps().whenComplete(_handleAppListFetched);
  }

  handleOpenDataApp(
      BuildContext context, Item plugin, AppState appState) async {
    var pluginId = plugin.get('id');

    RouteNavigator.navigateTo(
        route: Routes.appConfigure,
        context: context,
        param: {'id': pluginId, 'tab': appState.toUrlString()});
  }

  Item newPluginRunItemFromPlugin(Item plugin) {
    Item pluginRun = Item(type: 'PluginRun');
    pluginRun.copyPropertiesFrom(plugin, _runProperties);

    pluginRun.setIdIfNotExists();
    pluginRun.properties['targetItemId'] = pluginRun.id;
    Map<String, dynamic> config = {'isMock': true};
    String? dataType = plugin.get('dataType');
    if (dataType != null) {
      config['item_type'] = dataType;
    }
    pluginRun.properties['config'] = jsonEncode(config);
    return pluginRun;
  }

  runDataAppPlugin(Item plugin) async {
    AppLogger.debug('Creating pluginRun: ${plugin.get("name")}');
    Item pluginRun = newPluginRunItemFromPlugin(plugin);
    Edge edge = Edge(source: plugin, target: pluginRun, name: 'run');
    _podService.bulkAction(createItems: [pluginRun], createEdges: [edge]);
    currentMemriProject =
        await _graphService.getProjectFromPlugin(id: plugin.id);
    AppLogger.debug("created PluginRun ${pluginRun.get("id")}");
    _setupCurrentPluginListener(plugin, pluginRun);
  }

  _setupCurrentPluginListener(Item plugin, Item pluginRun) {
    var stream = _pluginRunStream(pluginRun.id);
    StreamSubscription<Item>? pluginRunSubscription;
    pluginRunSubscription = stream.listen((run) async {
      String? status = run.get('status');
      AppLogger.debug('status: $status');
      if (runIsDone(status)) {
        AppLogger.debug('Run done');
        // Stop plugin
        pluginRunSubscription?.cancel();
      }
      currentRunLog = await getRunLog(pluginRun);

      notifyListeners();
    });

    AppLogger.debug("created running plugin ${plugin.get("name")}");
  }

  void updateProjectSettings(
      RunningPlugin? runningPlugin, String? name, String? description) {
    if (runningPlugin == null) {
      return;
    }
    var plugin = runningPlugin.plugin;
    if (name != null && name.isNotEmpty) {
      plugin.properties['name'] = name;
    }
    // TODO error for empty name
    plugin.properties['pluginDescription'] = description;
    _podService.updateItem(item: plugin);
    notifyListeners();
  }

  _handleAppListFetching() {
    appListState = AppListState.fetching;
    // notifyListeners();
  }

  _handleAppListFetched() {
    appListState = AppListState.fetched;
    notifyListeners();
  }

  /// Inbox
  int selectedChannel = -1;
  List<Item> chats = [];
  List<Item> activeChatMessages = [];
  bool _loadingMessages = false;
  bool _loadingMoreMessages = false;
  bool _noChatRoomsFound = false;
  Map<String, String> chatImageSha = {};
  Map<String, ImageProvider> chatImageProviders = {};
  Map<String, String> messageImageSha = {};
  Map<String, String> tweetImageSha = {};
  Map<String, ImageProvider> messageImageProviders = {};
  Map<String, ImageProvider> tweetImageProviders = {};

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? chatStreamSubscription;

  // ignore: cancel_subscriptions
  StreamSubscription<List<Item>>? messageStreamSubscription;
  Stream<List<Item>>? chatStream, messageStream;
  String? selectedChatRoomID;
  ScrollController msgScrollController = ScrollController();

  bool get noChatRoomsFound => _noChatRoomsFound;

  bool get loadingMessages => _loadingMessages;

  bool get loadingMoreMessages => _loadingMoreMessages;

  Stream<List<Item>> gqlInboxChatsStream() async* {
    while (true) {
      await Future.delayed(Duration(milliseconds: 1500));
      List<Item> data = await _graphService.getInboxChats();
      _noChatRoomsFound = data.isEmpty;
      notifyListeners();
      if (data.isNotEmpty) {
        yield data;
      }
    }
  }

  Stream<List<Item>> gqlInboxMessagesStream(
      String chatID, int limit, int offset) async* {
    while (true) {
      await Future.delayed(Duration(milliseconds: 1500));
      List<Item> data =
          await _graphService.getInboxMessages(chatID, limit, offset);
      if (data.isNotEmpty) {
        yield data;
      }
    }
  }

  Stream<List<Item>> gqlStreamTwitterFeed() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      List<Item> data = await _graphService.getTwitterFeed();
      if (data.isNotEmpty) {
        // if (activeFeedItem == null) activeFeedItem = items[0];
        yield data;
      }
    }
  }

  Stream<List<Item>>? appContentStream;
  StreamSubscription<Item>? appContentSubscription;
  List<Item> appContent = [];

  void getChats() async {
    chatStream = gqlInboxChatsStream();
    chatStreamSubscription = chatStream!.listen((res) async {
      // for (var item in res) {
      //   print(item.edgeItem("~messageChannel")?.get("dateSent"));
      // }
      sortByEdgeProperty(
          items: res,
          edge: '~messageChannel',
          property: 'dateSent',
          order: 'desc');
      for (var item in res) {
        var itemID = item.get('id');
        String? fileID =
            item.edgeItem('photo')?.edgeItem('file')?.get('sha256')?.toString();
        if (fileID != null &&
            (!chatImageSha.containsKey(itemID) ||
                (chatImageSha.containsKey(itemID) &&
                    chatImageSha[itemID] != fileID))) {
          chatImageSha[itemID] = fileID;
          chatImageProviders[itemID] = MemoryImage(await downloadPhoto(fileID));
        }
      }
      chats = res;
      notifyListeners();
    });
  }

  stopAppContentStream() {
    appContentSubscription?.cancel();
    appContentStream = null;
  }

  void setActiveChat({required String chatID, limit = 100, offset = 0}) async {
    _loadingMessages = true;
    notifyListeners();
    if (kDebugMode) {
      AppLogger.debug('offset: $offset');
    }
    if (offset == 0) {
      if (selectedChatRoomID == chatID) return;

      selectedChatRoomID = chatID;
      // Reset active messages while waiting for query
      if (messageStreamSubscription != null) {
        messageStreamSubscription!.cancel();
      }
      messageStream = null;
      activeChatMessages = [];
      messageImageSha = {};
      messageImageProviders = {};
    }

    List<Item> messages = [];

    if (offset == 0) {
      messageStream = gqlInboxMessagesStream(chatID, limit, offset);

      messageStreamSubscription = messageStream!.listen((result) async {
        if (result.isNotEmpty) {
          _loadingMessages = true;
          notifyListeners();
          messages = result[0].getEdgeTargets('~messageChannel') ?? [];
          bool exist = false;
          for (Item message in messages) {
            exist = false;
            for (Item activeMessage in activeChatMessages) {
              if (activeMessage.get('id') == message.get('id')) {
                exist = true;
                break;
              }
            }
            if (!exist) {
              activeChatMessages.add(message);
            }
          }
          await updateMessageImage(activeChatMessages);
          _loadingMessages = false;
          notifyListeners();
        }
      });
    } else {
      _loadingMessages = false;
      notifyListeners();
      List<Item> result =
          await _graphService.getInboxMessages(chatID, limit, offset);
      if (result.isNotEmpty) {
        messages = result[0].getEdgeTargets('~messageChannel') ?? [];
        if (messages.isEmpty) {
          _loadingMoreMessages = false;
          notifyListeners();
          return;
        }
        await updateMessageImage(messages);
        activeChatMessages.addAll(messages);
      }
      _loadingMoreMessages = false;
      notifyListeners();
    }
  }

  Future<void> updateMessageImage(List<Item> messages) async {
    for (var item in messages) {
      var itemID = item.get('id');
      if (messageImageProviders[itemID] != null) continue;
      String? fileID = item
          .edgeItem('sender')!
          .edgeItem('profilePicture')
          ?.edgeItem('file')
          ?.get('sha256')
          ?.toString();
      if (fileID != null &&
          (!messageImageSha.containsKey(itemID) ||
              (messageImageSha.containsKey(itemID) &&
                  messageImageSha[itemID] != fileID))) {
        messageImageSha[itemID] = fileID;
        messageImageProviders[itemID] =
            MemoryImage(await downloadPhoto(fileID));
      }
    }
  }

  void sortByEdgeProperty(
      {required List<Item> items,
      required String edge,
      required String property,
      required String order}) {
    if (order == 'asc') {
      items.sort((a, b) {
        if (a.edgeItem(edge)?.get(property) == null) {
          return 1;
        } else if (b.edgeItem(edge)?.get(property) == null) {
          return -1;
        }
        return a
            .edgeItem(edge)!
            .get(property)
            .compareTo(b.edgeItem(edge)!.get(property));
      });
    } else {
      items.sort((b, a) {
        if (a.edgeItem(edge)?.get(property) == null) {
          return 1;
        } else if (b.edgeItem(edge)?.get(property) == null) {
          return -1;
        }
        return a
            .edgeItem(edge)!
            .get(property)
            .compareTo(b.edgeItem(edge)!.get(property));
      });
    }
  }

  void handleOnTapSelectedChatRoom(int index) {
    _loadingMessages = true;
    notifyListeners();
    selectedChannel = index;
    String? chatID = chats[index].get('id')?.toString();

    setActiveChat(chatID: chatID!);
    msgScrollController = ScrollController();
    msgScrollController.addListener(msgScrollControllerListener);
    notifyListeners();
  }

  void onVisibilityChanged(VisibilityInfo info) {
    var visiblePercentage = info.visibleFraction * 100;
    if (visiblePercentage < 20) {
      if (chatStreamSubscription != null && !chatStreamSubscription!.isPaused) {
        chatStreamSubscription!.pause();
      }
      if (messageStreamSubscription != null &&
          !messageStreamSubscription!.isPaused) {
        messageStreamSubscription!.pause();
      }
    } else {
      if (chatStreamSubscription != null && chatStreamSubscription!.isPaused) {
        chatStreamSubscription!.resume();
      }
      if (messageStreamSubscription != null &&
          !messageStreamSubscription!.isPaused) {
        messageStreamSubscription!.resume();
      }
    }
  }

  Future<String> getDataTypeForPlugin(String id) async {
    String? res = (await _podService.getItem(id: id))?.get('dataType');
    if (res == null) {
      throw Exception('no plugin found for id');
    }
    return res;
  }

  void loadMoreMessage() {
    _loadingMoreMessages = true;
    notifyListeners();
    setActiveChat(
      chatID: selectedChatRoomID!,
      offset: activeChatMessages.length,
    );
  }

  void msgScrollControllerListener() {
    // take into consideration reversed order of chat messages ListView
    if (msgScrollController.offset >=
            msgScrollController.position.maxScrollExtent &&
        !msgScrollController.position.outOfRange) {
      loadMoreMessage();
    }
  }

  void disposeStreams() {
    chatStream = null;
    messageStream = null;
    if (chatStreamSubscription != null) chatStreamSubscription!.cancel();
    if (messageStreamSubscription != null) messageStreamSubscription!.cancel();
  }

  void getFeed() async {
    feedStreamSubscription = gqlStreamTwitterFeed().listen((res) {
      feed = res;
      setTwitterImages();
      notifyListeners();
    });
  }

  void setTwitterImages() async {
    for (var item in feed) {
      var itemID = item.get('id');
      if (tweetImageProviders[itemID] != null) continue;
      String? fileID = item
          .edgeItem('author')
          ?.edgeItem('profilePicture')
          ?.edgeItem('file')
          ?.get('sha256')
          ?.toString();
      if (fileID != null &&
          (!tweetImageSha.containsKey(itemID) ||
              (tweetImageSha.containsKey(itemID) &&
                  tweetImageSha[itemID] != fileID))) {
        tweetImageSha[itemID] = fileID;
        tweetImageProviders[itemID] = MemoryImage(await downloadPhoto(fileID));
      }
    }
  }

  Future<void> setInitialItems() async {
    // set items if we already ran the plugin in the past
    String? dataSource = mockItemDataSourceForGit;
    if (dataSource == null) {
      debugPrint('Could not load initial items, no gitlab id set');
    } else {
      currentDataAppItems = await loadAppContent(dataSource, currentDataType!);
    }
  }

  Future<void> setLatestPluginRunAsCurrent(String containerImage) async {
    currentPluginRun =
        await _graphService.getPluginRunsWithContainerImage(containerImage);
  }

  Future<List<Item>> getPluginRunsForPlugin(String pluginId) async =>
      await _graphService.getPluginRunsForPlugin(pluginId);

  Future<Item?> getPlugin(String pluginId) async =>
      await _podService.getItem(id: pluginId);
}
