// ignore_for_file: cancel_subscriptions

import 'dart:async';
import 'dart:convert';
import 'dart:html' as html;
import 'dart:js' as js;
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/core/models/item.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/core/services/gitlab_service.dart';
import 'package:memri/core/services/graph_service.dart';
import 'package:memri/core/services/oauth_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';

class OAuthProvider with ChangeNotifier {
  final SharedPreferences _prefs;
  final PodService _podService;
  final OAuthService _oAuthService;
  final GraphService _graphService;
  final GitlabService _gitlabService;

  String? gitlabOauthUrl;

  OAuthProvider(
    this._prefs,
    this._podService,
    this._oAuthService,
    this._graphService,
    this._gitlabService,
  );

  String responseMessage = '';
  String errorMessage = 'Oops, something went wrong...!\nplease try again';
  String service = '';

  String? whatsappStatus;
  Uint8List? whatsappQrSvg;

  StreamSubscription<Item>? _whatsappOAuthStreamSubscription;
  StreamSubscription<dynamic>? _whatsappQrCodeStreamSubscription;
  Stream<Item>? _whatsappOAuthStream;
  Stream<dynamic>? _whatsappQrCodeStream;

  bool gitlabAuthFailed = false;
  StreamSubscription<void>? _gitlabOAuthAccessTokenStreamSubscription;
  Stream<void>? _gitlabOAuthStream;

  Map<DataSource, bool> oAuthCompleted = {};

  String get oauthTokenSecret => _prefs.getString('oauthTokenSecret') ?? '';

  String get oauthCodeVerifier => _prefs.getString('oauthCodeVerifier') ?? '';

  set oauthTokenSecret(String val) => _prefs.setString('oauthTokenSecret', val);

  set oauthCodeVerifier(String val) =>
      _prefs.setString('oauthCodeVerifier', val);

  bool oAuthCompletedByName(DataSource dataSource) {
    if (oAuthCompleted.keys.isEmpty) return false;
    return oAuthCompleted[dataSource] ?? false;
  }

  void init() {
    var uri = Uri.parse(Uri.base.toString());
    var state = uri.queryParameters['state'];
    if (state != null && state.contains('twitter')) {
      service = 'twitter';
      handleTwitterOAuth(uri.queryParameters);
    } else {
      var code = uri.queryParameters['code'];
      service = 'gitlab';
      if (code == null) {
        gitlabAuthFailed = true;
        responseMessage = errorMessage;
        AppLogger.info('No code provided');
      } else {
        _requestGitlabAccessTokens(code);
      }
    }
  }

  Future<dynamic> _requestGitlabAccessTokens(String code) async {
    String? accessToken = await _podService.oauth2Authorize(
        code, _gitlabService.redirectUrl, 'gitlab');
    debugPrint(accessToken);

    //success
    if (accessToken != null) {
      finishOAuth('GitLab');
    }
    //failure
    else {
      responseMessage = 'Oops, something went wrong. please try again';
      notifyListeners();
      // AppLogger.err(response.toString());
    }
  }

  Future<Item?> getLatestOAuthItem(String service) async {
    var items = await _graphService.getOAuthFlowItems(service);
    if (items.isNotEmpty) {
      items.sort((a, b) =>
          (a.get('dateCreated') as num).compareTo(b.get('dateCreated') as num));
      return items.last;
    } else {
      return null;
    }
  }

  void createOAuthItem(
      String service, String oauthToken, String oauthTokenSecret,
      {String? expiresIn, String? createdAt}) {
    Item item = Item.fromJson({
      'type': 'OauthFlow',
      'service': service,
      'accessToken': oauthToken,
      'accessTokenSecret': oauthTokenSecret,
    });
    if (expiresIn != null) {
      item.properties['expiresIn'] = expiresIn;
    }
    if (createdAt != null) {
      item.properties['createdAt'] = createdAt;
    }

    _podService.createItem(item: item);
  }

  setGitlabOauthUrl() async {
    List<String> scopes = ['api', 'read_repository', 'write_repository'];
    gitlabOauthUrl = await _podService.oauth2url(
        _gitlabService.redirectUrl, scopes, 'gitlab');
  }

  openOAuthWindow() {
    // safari only allows popup when a synchronous function is directly called on button click. Dont complicate this function
    html.window
        .open(gitlabOauthUrl!, '_blank', 'location=yes, height=800,width=600');
  }

  Future<Map<String, dynamic>> _requestOauthAccessToken(String oauthVerifier,
          String oauthToken, String oauthTokenSecret) async =>
      await _oAuthService.oauthAccessToken(
          oauthVerifier, oauthToken, oauthTokenSecret);

  _setOAuthCompleted(bool val) async {
    await _prefs.setBool('oauthCompleted', val);
    notifyListeners();
  }

  /// GITLAB
  bool _gitlabAuthTestSuccessful = false;

  bool get gitlabAuthTestSuccessful => _gitlabAuthTestSuccessful;

  bool get gitlabOAuthExist =>
      _gitlabService.gitlabAccessToken != null && _gitlabAuthTestSuccessful;

  _gitlabTestAuth() async {
    await _gitlabService.getUser().then((value) {
      _gitlabAuthTestSuccessful = true;
    }).catchError((error) {
      // TODO: Set error for user?
      _gitlabAuthTestSuccessful = false;
      _gitlabService.gitlabAccessToken = null;
    });
    notifyListeners();
  }

  Future<void> initGitlabOAuth() async {
    setGitlabOauthUrl();
    await setGitlabAccesTokenFromPod();
    if (_gitlabService.gitlabAccessToken == null) {
      // start a listener for accessToken, if it finds one, set in the gitlab api and test auth
      _gitlabOAuthStream = _gitlabLatestOauthItemStream();
      _gitlabOAuthAccessTokenStreamSubscription =
          _gitlabOAuthStream!.listen((item) async {
        await _gitlabTestAuth();
        cancelAndDeleteGitlabStream();
        notifyListeners();
      });
    } else {
      // tests if item can be used for auth
      await _gitlabTestAuth();
    }
  }

  Stream<String> _gitlabLatestOauthItemStream() async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      await setGitlabAccesTokenFromPod();
      if (_gitlabService.gitlabAccessToken != null) {
        yield _gitlabService.gitlabAccessToken!;
      }
    }
  }

  Future<void> setGitlabAccesTokenFromPod() async {
    try {
      _gitlabService.gitlabAccessToken ??=
          await _podService.oauth2GetAccessToken('gitlab');
    } catch (e) {
      debugPrint('Failed to get accessToken');
    }
  }

  cancelAndDeleteGitlabStream() {
    _gitlabOAuthStream = null;
    if (_gitlabOAuthAccessTokenStreamSubscription != null) {
      _gitlabOAuthAccessTokenStreamSubscription!.cancel();
    }
  }

  /// Twitter
  ///

  String? twitterOauthToken;
  String? twitterOauthTokenSecret;

  setTwitterOauthToken() async {
    String redirectUrl = '${_gitlabService.endpoint}/oauth?state=twitter';
    var res = await _oAuthService.oauthRequestToken('twitter', redirectUrl);
    twitterOauthToken = res['oauth_token'];
    twitterOauthTokenSecret = res['oauth_token_secret'];
  }

  void initTwitterOAuth(BuildContext context) {
    _setOAuthCompleted(false);

    final queryParameters = {'oauth_token': twitterOauthToken!};

    // we are using this in the redirect screen
    _prefs.setString('oauthTokenSecret', twitterOauthTokenSecret!);

    var url = Uri.https('api.twitter.com', '/oauth/authorize', queryParameters);

    html.window
        .open(url.toString(), '_blank', 'location=yes, height=800,width=600');

    Timer.periodic(Duration(seconds: 1), (timer) async {
      _prefs.reload();
      notifyListeners();

      oAuthCompleted[DataSource.twitter] =
          _prefs.getBool('oauthCompleted') ?? false;

      if (oAuthCompleted[DataSource.twitter]!) {
        AppLogger.info('oauth completed');
        Item item = await _podService.startImporter('TwitterPlugin',
            'twitter.plugin', app.settings.twitterContainer, 'twitter');
        var status = item.get('status');
        notifyListeners();
        timer.cancel();
        if (status == 'error') {
          ///
          /// TODO
          ///
        } else {
          String id = item.id;
          RouteNavigator.navigateTo(
            context: context,
            route: Routes.importerDownloading,
            param: {'id': id, 'name': 'twitter'},
            replace: true,
          );
        }
      }
    });
  }

  void handleTwitterOAuth(Map<String, String> previousQueryParameters) async {
    var resBody = await _requestOauthAccessToken(
        previousQueryParameters['oauth_verifier']!,
        previousQueryParameters['oauth_token']!,
        oauthTokenSecret);

    //success
    if (resBody['oauth_token'] != null) {
      String oauthToken = resBody['oauth_token']!;
      String oauthTokenSecret = resBody['oauth_token_secret']!;
      createOAuthItem('twitter', oauthToken, oauthTokenSecret);
      AppLogger.info('setting');

      _setOAuthCompleted(true);
      finishOAuth('Twitter');
    }
    //failure
    else {
      responseMessage = 'Oops, something went wrong. please try again';
      notifyListeners();
      AppLogger.err(resBody.toString());
    }
    // close screen
  }

  void finishOAuth(String title) async {
    for (var i in [3, 2, 1]) {
      await Future.delayed(Duration(seconds: 1));
      responseMessage =
          'Successfully authenticated $title\nRedirecting you back in $i seconds';
      notifyListeners();
    }
    await Future.delayed(Duration(seconds: 1));
    js.context.callMethod('close');
  }

  /// WhatsApp

  void initWhatsappAuth(BuildContext context) async {
    // TODO, what are we doing with things from the default database?

    try {
      Item item = await _podService.startImporter('WhatsappPlugin',
          'whatsapp.plugin', app.settings.whatsappContainer, 'whatsapp');
      String id = item.id;
      service = 'whatsapp';
      _whatsappOAuthStream ??= _whatsappItemStream(id);
      _whatsappOAuthStreamSubscription =
          _whatsappOAuthStream!.listen((item) async {
        cancelWhatsappStreamSubs();
        String? webserverUrl = item.get('webserverUrl');
        if (webserverUrl != null) {
          _whatsappQrCodeStream ??= _whatsappQrStream(webserverUrl);
          _whatsappQrCodeStreamSubscription =
              _whatsappQrCodeStream!.listen((dynamic res) async {
            if (res != null) {
              oAuthCompleted[DataSource.whatsapp] = res['authenticated'];
              if (oAuthCompleted[DataSource.whatsapp]!) {
                RouteNavigator.navigateTo(
                  context: context,
                  route: Routes.importerDownloading,
                  param: {'id': id, 'name': 'whatsapp'},
                  replace: true,
                );
              } else {
                String? svg = res['qr'];
                if (svg != null) {
                  whatsappQrSvg = base64.decode(svg.split(',').last);
                  notifyListeners();
                }
              }
            }
          });

          var status = item.get('status');
          if (status != whatsappStatus) {
            AnalyticsService().logImporterStatus(status);
          }
          whatsappStatus = status;
          AppLogger.info(whatsappStatus);

          notifyListeners();
        }
      });
    } catch (error) {
      AppLogger.err('Error starting plugin: $error');
    }
  }

  void cancelWhatsappStreamSubs() {
    _whatsappOAuthStream = null;
    if (_whatsappOAuthStreamSubscription != null) {
      _whatsappOAuthStreamSubscription!.cancel();
    }
  }

  void cancelWhatsappQrCodeStreamSubs() {
    _whatsappQrCodeStream = null;
    if (_whatsappQrCodeStreamSubscription != null) {
      _whatsappQrCodeStreamSubscription!.cancel();
    }
  }

  Stream<Item> _whatsappItemStream(String id) async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      Item? res;
      Item? pluginRunItem = await _podService.getItem(id: id);
      if (pluginRunItem != null) {
        res = pluginRunItem;
      }

      if (res != null) {
        yield res;
        if (res.get('webserverUrl') != null) {
          cancelWhatsappStreamSubs();
        }
      }
    }
  }

  Stream<dynamic> _whatsappQrStream(String webserverUrl) async* {
    while (true) {
      await Future.delayed(Duration(seconds: 1));
      dynamic res = await _oAuthService.getQrCodeSvg(webserverUrl);
      if (res != null) {
        yield res;
        if (res['authenticated']) {
          cancelWhatsappQrCodeStreamSubs();
        }
      }
    }
  }

  ///

  void reset() {
    cancelWhatsappStreamSubs();
    cancelWhatsappQrCodeStreamSubs();
    _gitlabOAuthStream = null;
    _gitlabService.gitlabAccessToken = null;
    whatsappQrSvg = null;
    whatsappStatus = null;
    oAuthCompleted = {};
    gitlabAuthFailed = false;
    service = '';
    responseMessage = '';
    errorMessage = '';
  }
}
