import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/core/services/pod_service.dart';
import 'package:memri/localization/generated/l10n.dart';
import 'package:memri/memri.dart';
import 'package:memri/providers/app_provider.dart';
import 'package:memri/providers/data_app_provider.dart';
import 'package:memri/providers/importer_provider.dart';
import 'package:memri/providers/oauth_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:shared_preferences/shared_preferences.dart';

enum AuthState {
  authentication,
  authenticating,
  authenticated,
  savedKeys,
  error,
}

enum DeveloperAuthState { devSignIn, devSignUp }

class PodProvider with ChangeNotifier {
  final AppProvider _appProvider;
  final PodService _podService;
  final AnalyticsService _analyticsService;
  final OAuthProvider _oAuthProvider;
  final SharedPreferences _prefs;

  final DataAppProvider _dataAppProvider;
  final ProjectProvider _projectProvider;
  final ImporterProvider _importerProvider;

  PodProvider(
    this._prefs,
    this._appProvider,
    this._podService,
    this._analyticsService,
    this._oAuthProvider,
    this._dataAppProvider,
    this._projectProvider,
    this._importerProvider,
  );

  AuthState state = AuthState.authentication;
  DeveloperAuthState devState = DeveloperAuthState.devSignUp;
  String errorMessage = '';
  bool _developerMode = false;
  bool _signupRefresh = true;

  String get ownerKey => _podService.podConfig.ownerKey;

  String get databaseKey => _podService.podConfig.databaseKey;

  bool get developerMode => _developerMode;

  Future<void> _authenticate(BuildContext context, String podUrl,
      String ownerKey, String dbKey) async {
    try {
      await _podService.authenticate(
          podAddress: podUrl, ownerKey: ownerKey, dbKey: dbKey);
      // TODO move this once not required on signup
      await _podService.createSchema();
      // ignore: use_build_context_synchronously
      _handleAuthenticated(context);
    } on Exception catch (_) {
      _appProvider.closeLoadingDialog(context);

      if (devState == DeveloperAuthState.devSignUp) {
        _handleError(S.current.account_login_wrong_keys_error);
      } else {
        /// TODO should handle different error types
        _handleError(S.current.account_login_general_error);
      }
    }
  }

  Future<void> login({
    required BuildContext context,
    required String ownerKey,
    required String dbKey,
    String? podAddress,
  }) async {
    if (ownerKey.isEmpty) {
      _handleError(S.current.account_login_empty_owner_key_error);
      return;
    } else if (dbKey.isEmpty) {
      _handleError(S.current.account_login_empty_database_key_error);
      return;
    } else if (podAddress != null && podAddress.length < 3) {
      _handleError(S.current.account_login_empty_pod_url_error);
      return;
    }
    _handleAuthenticating(context);
    var podUrl = podAddress ?? app.settings.defaultPodUrl;

    _analyticsService.logSignIn(ownerKey, podAddress != null);
    await _authenticate(context, podUrl, ownerKey, dbKey);
    RouteNavigator.navigateTo(
        context: context, route: Routes.workspace, clearStack: true);
  }

  Future<void> signUp(BuildContext context,
      {String? podAddress, bool refresh = true}) async {
    _signupRefresh = refresh;
    _handleAuthenticating(context);
    var podUrl = podAddress ?? app.settings.defaultPodUrl;
    if (podUrl.length < 3) {
      _handleError(S.current.account_login_empty_pod_url_error);
      _signupRefresh = true;
      return;
    }
    String ownerKey = app.userAccessKey != null && app.userAccessKey!.isNotEmpty
        ? app.userAccessKey!
        : _podService.generateCryptoStrongKey();
    String dbKey = _podService.generateCryptoStrongKey();

    _analyticsService.logSignUp(ownerKey, podAddress != null);
    await _authenticate(context, podUrl, ownerKey, dbKey);
    _signupRefresh = true;
    RouteNavigator.navigateTo(
        context: context, route: Routes.saveKeys, clearStack: true);
  }

  Future<void> signOut(BuildContext context) async {
    await _prefs.clear();
    _analyticsService.reset();
    _oAuthProvider.reset();
    _appProvider.podVersion = null;
    providerKey = app.randomUuid();
    _dataAppProvider.reset();
    _projectProvider.reset();
    _importerProvider.reset();
    RouteNavigator.navigateTo(
        context: context, route: Routes.onboarding, clearStack: true);
  }

  void openLoginScreen(BuildContext context, {bool developerMode = false}) {
    _developerMode = developerMode;
    RouteNavigator.navigateTo(
        context: context,
        route: developerMode ? Routes.loginDev : Routes.login);
  }

  void copyKeysToClipboard({String? text}) {
    Clipboard.setData(ClipboardData(
        text: text ??
            '${S.current.login_key}: $ownerKey\n${S.current.password_key}: $databaseKey'));
    _handleSavedKeys();
  }

  bool _keyVisibility = false;

  bool get keyVisibility => _keyVisibility;

  void handleKeysMessageVisibility(bool show) {
    _keyVisibility = show;
    notifyListeners();
  }

  void finishAuthentication(BuildContext context) {
    WidgetsBinding.instance!.addPostFrameCallback((_) =>
        RouteNavigator.navigateTo(
            context: context, route: Routes.workspace, clearStack: true));
  }

  void _handleAuthenticating(BuildContext context) {
    state = AuthState.authenticating;
    errorMessage = '';
    if (_signupRefresh) {
      notifyListeners();
      _appProvider.showLoadingDialog(context,
          message:
              '${S.current.authenticating}\n${S.current.authenticating_message}');
    }
  }

  void _handleAuthenticated(BuildContext context) {
    state = AuthState.authenticated;
    errorMessage = '';
    if (_signupRefresh) notifyListeners();
    _appProvider.closeLoadingDialog(context);
  }

  void _handleSavedKeys() {
    state = AuthState.savedKeys;
    errorMessage = '';
    if (_signupRefresh) notifyListeners();
  }

  void _handleError(e) {
    state = AuthState.error;
    errorMessage = e.toString();
    notifyListeners();
  }

  void updateDevStateToSignIn() {
    devState = DeveloperAuthState.devSignIn;
    state = AuthState.authenticated;
    errorMessage = '';
    notifyListeners();
  }

  void updateDevStateToSignUp() {
    devState = DeveloperAuthState.devSignUp;
    state = AuthState.authenticated;
    errorMessage = '';
    notifyListeners();
  }

  void initWorkspace() {
    if (_podService.podConfig.ownerKey.isNotEmpty) {
      state = AuthState.authenticated;
    }
  }
}
