import 'package:flutter/material.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/providers/data_app_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/widgets/loading_indicator.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:memri/widgets/screen_error.dart';
import 'package:provider/provider.dart';
import 'package:visibility_detector/visibility_detector.dart';

class AppsInboxScreen extends StatefulWidget {
  final bool showMainNavigation;
  final String? importer;

  const AppsInboxScreen(
      {Key? key, this.showMainNavigation = true, this.importer})
      : super(key: key);

  @override
  AppsInboxScreenState createState() => AppsInboxScreenState();
}

class AppsInboxScreenState extends State<AppsInboxScreen> {
  late final _dataAppProvider =
      Provider.of<DataAppProvider>(context, listen: false);

  @override
  void initState() {
    _dataAppProvider.getChats();
    super.initState();
  }

  @override
  void dispose() {
    _dataAppProvider.disposeStreams();
    _dataAppProvider.msgScrollController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.apps,
      backgroundColor: app.colors.white,
      child: Consumer<DataAppProvider>(builder: (context, provider, _) {
        if (provider.noChatRoomsFound && provider.chats.isEmpty) {
          return Padding(
            padding: const EdgeInsets.symmetric(horizontal: 30.0),
            child: ScreenError(
              title: 'Inbox Error',
              message:
                  "Sorry, we couldn't find any results. Please, make sure you connect at least one data source importer.",
              buttonLabel: 'Back to apps',
            ),
          );
        }
        return provider.chats.isEmpty && !provider.noChatRoomsFound
            ? LoadingIndicator(message: 'Fetching your chat rooms...!')
            : VisibilityDetector(
                key: Key('inbox'),
                onVisibilityChanged: provider.onVisibilityChanged,
                child: Container(
                  width: MediaQuery.of(context).size.width,
                  color: app.colors.white,
                  child: Row(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: [
                      Expanded(flex: 1, child: _buildChatRoomsBody(provider)),
                      VerticalDivider(width: 0),
                      Expanded(flex: 2, child: _buildMessagesBody(provider)),
                    ],
                  ),
                ),
              );
      }),
    );
  }

  Widget _buildChatRoomsBody(DataAppProvider provider) {
    return ListView.builder(
        controller: ScrollController(),
        itemCount: provider.chats.length,
        padding: EdgeInsets.only(top: 40),
        shrinkWrap: true,
        itemBuilder: (context, index) {
          String name = provider.chats[index].get('name') ?? ' ';
          var image =
              provider.chatImageProviders[provider.chats[index].get('id')];
          return Container(
            color: provider.selectedChannel == index
                ? app.colors.greyLight
                : app.colors.white,
            padding: EdgeInsets.symmetric(vertical: 5, horizontal: 20),
            child: ListTile(
              onTap: () => provider.handleOnTapSelectedChatRoom(index),
              leading: _buildAvatar(image, name),
              horizontalTitleGap: 15,
              title: Text(name),
              // trailing: IconButton(
              //   onPressed: () {},
              //   icon: Icon(
              //     Icons.more_horiz,
              //     color: Color(0xFFDFDEDE),
              //   ),
              // ),
            ),
          );
        });
  }

  Widget _buildMessagesBody(DataAppProvider provider) {
    if (provider.loadingMessages) {
      return LoadingIndicator(message: 'Fetching your messages...!');
    }
    if (provider.selectedChatRoomID == null) {
      return Center(
          child: Text(
        'Select a chat room to see its messages',
        style: AppStyles.bodyText1,
      ));
    }
    if (provider.activeChatMessages.isEmpty && provider.noChatRoomsFound) {
      return SizedBox();
    }
    if (provider.activeChatMessages.isEmpty) {
      return Center(
          child: Text(
        'No message found',
        style: AppStyles.bodyText1,
      ));
    }

    return Stack(
      children: [
        if (provider.loadingMoreMessages)
          Positioned(
            top: 0,
            right: 0,
            left: 0,
            height: 60,
            child: Column(
              children: [
                Container(
                  color: Colors.transparent,
                  padding: EdgeInsets.all(8),
                  child: CircularProgressIndicator(
                    backgroundColor: Colors.transparent,
                    color: app.colors.brandOrange,
                  ),
                ),
              ],
            ),
          ),
        Padding(
          padding: EdgeInsets.only(top: provider.loadingMoreMessages ? 60 : 0),
          child: ListView.builder(
            controller: _dataAppProvider.msgScrollController,
            reverse: true,
            itemCount: provider.activeChatMessages.length,
            shrinkWrap: true,
            itemBuilder: (context, index) {
              String name = provider.activeChatMessages[index]
                      .edgeItem('sender')!
                      .get('displayName') ??
                  ' ';
              var image = provider.messageImageProviders[
                  provider.activeChatMessages[index].get('id')];

              return ListTile(
                contentPadding: EdgeInsets.symmetric(horizontal: 30),
                minLeadingWidth: 15,
                leading: _buildAvatar(image, name),
                title: Builder(builder: (context) {
                  var value = provider.activeChatMessages[index]
                      .get('content')
                      ?.toString();
                  if (value == null) {
                    return Text('');
                  } else {
                    bool byMe = provider.activeChatMessages[index]
                            .edgeItem('sender')
                            ?.get('isMe') ??
                        false;
                    return Column(
                      crossAxisAlignment: CrossAxisAlignment.start,
                      children: [
                        Container(
                          padding: EdgeInsets.symmetric(
                              vertical: 10, horizontal: 15),
                          decoration: BoxDecoration(
                            color: byMe
                                ? app.colors.brandViolet
                                : app.colors.brandWhite,
                            borderRadius: BorderRadius.only(
                              topRight: Radius.circular(10),
                              bottomRight: Radius.circular(10),
                              bottomLeft: Radius.circular(byMe ? 10 : 0),
                              topLeft: Radius.circular(byMe ? 0 : 10),
                            ),
                          ),
                          child: Text(
                            value,
                            style: AppStyles.bodyText1.copyWith(
                              color: byMe
                                  ? app.colors.brandWhite
                                  : app.colors.brandBlack,
                            ),
                          ),
                        ),
                      ],
                    );
                  }
                }),
              );
            },
          ),
        ),
      ],
    );
  }

  Widget _buildAvatar(ImageProvider? image, String name) {
    return CircleAvatar(
      radius: 16.0,
      backgroundImage: image,
      backgroundColor: app.colors.brandOrange,
      child: image == null
          ? Text(
              name.substring(0, 1),
              style: AppStyles.buttonLabel.copyWith(color: app.colors.white),
            )
          : null,
    );
  }
}
