import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/providers/data_app_provider.dart';
import 'package:memri/providers/importer_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/widgets/data_app_tile.dart';
import 'package:memri/widgets/empty.dart';
import 'package:memri/widgets/loading_indicator.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class AppsScreen extends StatefulWidget {
  final bool showMainNavigation;

  const AppsScreen({Key? key, this.showMainNavigation = true})
      : super(key: key);

  @override
  AppsScreenState createState() => AppsScreenState();
}

class AppsScreenState extends State<AppsScreen> {
  @override
  void initState() {
    Provider.of<DataAppProvider>(context, listen: false).fetchDataApps();
    Provider.of<ImporterProvider>(context, listen: false).fetchImportersData();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.apps,
      backgroundColor: app.colors.white,
      child: SingleChildScrollView(
        child: Container(
          padding: const EdgeInsets.all(30.0),
          child: Row(
            children: [
              Expanded(
                child: Consumer<DataAppProvider>(
                    builder: (context, provider, _) =>
                        Consumer<ImporterProvider>(
                          builder: (context, importerProvider, __) {
                            bool whatsappExist = importerProvider
                                    .importerDataInPod[DataSource.whatsapp] ??
                                false;
                            bool twitterExist = importerProvider
                                    .importerDataInPod[DataSource.twitter] ??
                                false;
                            return Column(
                              crossAxisAlignment: CrossAxisAlignment.start,
                              children: [
                                whatsappExist ? _buildInboxTile() : SizedBox(),
                                SizedBox(
                                    height:
                                        twitterExist && whatsappExist ? 15 : 0),
                                twitterExist ? _buildFeedTile() : SizedBox(),
                                if (provider.appListState ==
                                    AppListState.fetching)
                                  Container(
                                    width: MediaQuery.of(context).size.width,
                                    padding: const EdgeInsets.symmetric(
                                        vertical: 60),
                                    child: Center(
                                        child: LoadingIndicator(
                                      height: 75,
                                      message: 'Fetching Data Apps...',
                                    )),
                                  )
                                else if (provider.appListState ==
                                    AppListState.fetched)
                                  provider.dataApps.isEmpty
                                      ? _buildEmptyState()
                                      : _buildAppListView(provider),
                              ],
                            );
                          },
                        )),
              ),
              Expanded(child: Empty()),
            ],
          ),
        ),
      ),
    );
  }

  Widget _buildFeedTile() => _buildTile(
        'Feeds',
        'This is your default feed. It displays all social media from connected data streams.',
        Routes.feed,
      );

  Widget _buildTile(String title, String description, String dest) {
    return InkWell(
      onTap: () => RouteNavigator.navigateTo(context: context, route: dest),
      child: Container(
        width: MediaQuery.of(context).size.width,
        color: Color(0xfff0f0f0),
        padding: EdgeInsets.all(20),
        child: Column(
          mainAxisSize: MainAxisSize.min,
          crossAxisAlignment: CrossAxisAlignment.start,
          children: [
            Text(title, style: AppStyles.headline2),
            SizedBox(height: 45),
            Text(
              description,
              style:
                  AppStyles.bodyText1.copyWith(color: app.colors.brandGreyText),
            ),
            SizedBox(height: 30),
            TextButton(
              style: ButtonStyle(
                  padding: MaterialStateProperty.all(EdgeInsets.all(0))),
              onPressed: () =>
                  RouteNavigator.navigateTo(context: context, route: dest),
              child: Text(
                'Open App',
                style:
                    AppStyles.buttonLabel.copyWith(color: app.colors.primary),
              ),
            ),
          ],
        ),
      ),
    );
  }

  Widget _buildInboxTile() {
    return _buildTile(
        'Inbox',
        'This is your default inbox app. It displays messages from all your connected data streams. You can send and receive messages directly within this app. You must connect at least one messenger or email importer to use this app.',
        Routes.inbox);
  }

  Widget _buildEmptyState() {
    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        SizedBox(height: 100),
        Text('Start building your first data app', style: AppStyles.headline2),
        SizedBox(height: 20),
        Text(
          'Memri allows to build, deploy and share ML apps on personal data.\n\nStart building an app by creating a new project.',
          style: AppStyles.bodyText1,
        ),
        SizedBox(height: 30),
        Consumer<ImporterProvider>(builder: (_, importerProvider, __) {
          return TextButton(
            onPressed: () => RouteNavigator.navigateTo(
                context: context,
                route: importerProvider.isAnyImportersActive
                    ? Routes.projectsCreate
                    : Routes.data),
            style: primaryButtonStyle,
            child: SizedBox(
              height: 36,
              child: Row(
                mainAxisSize: MainAxisSize.min,
                children: [
                  Text(importerProvider.isAnyImportersActive
                      ? 'Create new project'
                      : 'Connect data'),
                  SizedBox(width: 10),
                  app.icons.plus(),
                ],
              ),
            ),
          );
        }),
      ],
    );
  }

  Widget _buildAppListView(DataAppProvider provider) {
    return Column(
      children: List.generate(
        provider.dataApps.length,
        (index) => Padding(
          padding: const EdgeInsets.only(top: 15),
          child: DataAppTile(dataApp: provider.dataApps[index]),
        ),
      ),
    );
  }
}
