import 'dart:js' as js;

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_logger.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/providers/oauth_provider.dart';
import 'package:memri/providers/pod_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/custom_loading_indicator.dart';
import 'package:memri/widgets/project/project_progress_bar.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class ProjectsAppCreateScreen extends StatefulWidget {
  final String id;

  const ProjectsAppCreateScreen({Key? key, required this.id}) : super(key: key);

  @override
  State<ProjectsAppCreateScreen> createState() =>
      _ProjectsAppCreateScreenState();
}

class _ProjectsAppCreateScreenState extends State<ProjectsAppCreateScreen> {
  late final pProvider = Provider.of<ProjectProvider>(context, listen: false);
  late String ownerKey;
  late String databaseKey;

  @override
  void initState() {
    var podProvider = Provider.of<PodProvider>(context, listen: false);
    ownerKey = podProvider.ownerKey;
    databaseKey = podProvider.databaseKey;
    _setupGitlab();
    super.initState();
  }

  Future<void> _setupGitlab() async {
    AppLogger.info('init gitlab');
    var oAuthProvider = Provider.of<OAuthProvider>(context, listen: false);
    await pProvider.setCurrentMemriProject(context, widget.id);
    await oAuthProvider.initGitlabOAuth();
    await pProvider.setGitlabProject();
    pProvider.setupProjectPackageStream();
    pProvider.setupProjectPipelineStream();
  }

  @override
  void dispose() {
    if (pProvider.projectPackageSubscription != null) {
      pProvider.projectPackageSubscription!.cancel();
    }
    if (pProvider.projectPipelineSubscription != null) {
      pProvider.projectPipelineSubscription!.cancel();
    }
    pProvider.projectPackageStream = null;
    pProvider.projectPipelineStream = null;
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.projects,
      child: ProjectProgressBar(currentStep: 2, child: _buildBody()),
    );
  }

  Widget _buildBody() {
    return SingleChildScrollView(
      child: Consumer<ProjectProvider>(
        builder: (context, provider, _) {
          return Container(
            color: app.colors.white,
            width: MediaQuery.of(context).size.width,
            padding: EdgeInsets.symmetric(horizontal: 30),
            child: Column(
              crossAxisAlignment: CrossAxisAlignment.start,
              children: [
                SizedBox(height: 20),
                if (!ResponsiveHelper(context).isSmallScreen)
                  SizedBox(height: 20),
                Text('Load a dataset, finetune and train a model',
                    style: AppStyles.headline1),
                SizedBox(height: 10),
                Text(
                    'You can either use our preconfigured template, with setup, testing and CI already set up, or use your preferred tools to do this.',
                    style: AppStyles.bodyText1),
                SizedBox(height: 60),
                Text('Train an ML model on Google Colab',
                    style: AppStyles.headline3),
                SizedBox(height: 10),
                Text(
                    'Google Colab allows you to load and train models faster. Just load the model into our template and follow the on-screen instructions.',
                    style: AppStyles.bodyText1),
                SizedBox(height: 20),
                TextButton(
                  onPressed: () {
                    AnalyticsService().logProjectGoogleColab();
                    js.context.callMethod('open', [app.settings.colabLink]);
                  },
                  style: provider.projectHasPackage
                      ? secondaryButtonStyle
                      : primaryButtonStyle,
                  child: SizedBox(
                    height: 36,
                    child: Row(
                      mainAxisSize: MainAxisSize.min,
                      children: [
                        if (!provider.projectHasPackage)
                          app.images.colab(color: app.colors.white),
                        if (!provider.projectHasPackage) SizedBox(width: 16),
                        Text(
                          provider.projectHasPackage
                              ? 'Your project on Google Colab'
                              : 'Open your project in Google Colab',
                          style: provider.projectHasPackage
                              ? AppStyles.buttonLabel
                                  .copyWith(color: app.colors.brandOrange)
                              : AppStyles.buttonLabel
                                  .copyWith(color: app.colors.greyBackGround),
                        )
                      ],
                    ),
                  ),
                ),
                SizedBox(height: 70),
                if (provider.projectHasPackage)
                  Column(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: [
                      Text('Well done!',
                          style: AppStyles.headline3
                              .copyWith(color: app.colors.brandGreen)),
                      SizedBox(height: 8),
                      Text(
                          'Your model has been pushed to your repository and is being deployed by the CI pipeline.',
                          style: AppStyles.headline3
                              .copyWith(color: app.colors.brandGreen)),
                      SizedBox(height: 30),
                      Row(
                        children: const [
                          Text('The process will take about:',
                              style: AppStyles.bodyText1),
                          SizedBox(width: 10),
                          Text('2min and 17s', style: AppStyles.bodyBold),
                        ],
                      ),
                      SizedBox(height: 5),
                      Row(
                        children: [
                          Text('Current status:', style: AppStyles.bodyText1),
                          SizedBox(width: 10),
                          for (var job in provider.projectPipelines)
                            Row(
                              children: [
                                InkWell(
                                  onTap: () => {
                                    js.context.callMethod('open', [job.url])
                                  },
                                  child: Container(
                                    height: 40,
                                    color: app.colors.greyBackGround,
                                    padding: EdgeInsets.all(8),
                                    child: Row(
                                      children: [
                                        Text(
                                          job.name,
                                          style: {
                                            'running': AppStyles.buttonLabel
                                                .copyWith(
                                                    color:
                                                        app.colors.brandYellow),
                                            'success': AppStyles.buttonLabel
                                                .copyWith(
                                                    color:
                                                        app.colors.brandGreen),
                                            'pending': AppStyles.buttonLabel
                                                .copyWith(
                                                    color: app
                                                        .colors.brandGreyText),
                                            'failed': AppStyles.buttonLabel
                                                .copyWith(
                                                    color:
                                                        app.colors.brandOrange),
                                          }[job.status],
                                        ),
                                        SizedBox(width: 8),
                                        if (job.status == 'running')
                                          CustomLoadingIndicator(
                                            color: app.colors.brandYellow,
                                            size: 30,
                                            speed: 0.8,
                                          )
                                        else if (job.status == 'pending')
                                          SizedBox.shrink()
                                        else if (job.status == 'failed')
                                          app.icons.x(
                                              color: app.colors.brandOrange,
                                              width: 30)
                                        else if (job.status == 'success')
                                          app.icons.checkmark(
                                              color: app.colors.brandGreen)
                                        else
                                          app.icons.checkmark()
                                      ],
                                    ),
                                  ),
                                ),
                                SizedBox(width: 10),
                              ],
                            )
                        ],
                      ),
                      SizedBox(height: 70),
                    ],
                  ),
                Container(
                  color: app.colors.greyBackGround,
                  padding: EdgeInsets.all(20),
                  child: Column(
                    crossAxisAlignment: CrossAxisAlignment.start,
                    children: [
                      Text('You will need this information:',
                          style: AppStyles.bodyBold),
                      SizedBox(height: 28),
                      DatasetInfoRow(
                        title: 'Your dataset name',
                        content: provider.getDataset()?.get('name') ?? '',
                        hide: false,
                      ),
                      SizedBox(height: 28),
                      DatasetInfoRow(
                        title: 'your owner key',
                        content: ownerKey,
                        hide: true,
                      ),
                      SizedBox(height: 28),
                      DatasetInfoRow(
                        title: 'your database key',
                        content: databaseKey,
                        hide: true,
                      ),
                      SizedBox(height: 28),
                      DatasetInfoRow(
                        title: 'your plugin gitlab project name',
                        content: provider.gitlabProject['path'] ?? '',
                        hide: false,
                      )
                    ],
                  ),
                ),
                SizedBox(height: 70),
                Wrap(
                  children: [
                    TextButton(
                      onPressed:
                          provider.projectCIPassed && provider.projectHasPackage
                              ? () => provider.handleCreateApp(context)
                              : null,
                      style: primaryButtonStyle.copyWith(
                        backgroundColor: MaterialStateProperty.all(
                            provider.projectCIPassed &&
                                    provider.projectHasPackage
                                ? app.colors.brandOrange
                                : app.colors.greyInactive),
                      ),
                      child: SizedBox(
                        height: 36,
                        child: Padding(
                          padding: const EdgeInsets.symmetric(horizontal: 10),
                          child: Row(
                            mainAxisSize: MainAxisSize.min,
                            children: [
                              Text('Next', style: AppStyles.buttonLabel),
                              SizedBox(width: 16),
                              app.icons.arrow()
                            ],
                          ),
                        ),
                      ),
                    ),
                    SizedBox(width: 10),
                    TextButton(
                        onPressed: () => RouteNavigator.navigateTo(
                              route: Routes.projectsLabelData,
                              context: context,
                              param: {'id': widget.id},
                              replace: true,
                            ),
                        style: secondaryButtonStyle,
                        child: SizedBox(
                          height: 36,
                          child: Padding(
                            padding: const EdgeInsets.symmetric(horizontal: 10),
                            child: Row(
                              mainAxisSize: MainAxisSize.min,
                              children: const [Text('Back to labelling')],
                            ),
                          ),
                        ))
                  ],
                ),
                SizedBox(height: 20),
                if (!ResponsiveHelper(context).isSmallScreen)
                  SizedBox(height: 20),
              ],
            ),
          );
        },
      ),
    );
  }
}

class DatasetInfoRow extends StatelessWidget {
  const DatasetInfoRow(
      {Key? key,
      required this.title,
      required this.content,
      required this.hide})
      : super(key: key);

  final String title;
  final String content;
  final bool hide;

  @override
  Widget build(BuildContext context) {
    return Row(
      children: [
        Expanded(
          flex: 2,
          child: Row(
            children: [
              Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                children: [
                  Text(
                    title.toUpperCase(),
                    style:
                        AppStyles.smallCaps.copyWith(color: Color(0xff828282)),
                  ),
                  SizedBox(height: 5),
                  (hide)
                      ? Text('**** **** **** **** **** ****',
                          style: AppStyles.code)
                      : SelectableText(content, style: AppStyles.code),
                ],
              ),
              Spacer(),
              InkWell(
                onTap: () => Clipboard.setData(ClipboardData(text: content)),
                child: app.icons.duplicate(color: app.colors.brandOrange),
              ),
            ],
          ),
        ),
        Expanded(child: Container())
      ],
    );
  }
}
