import 'package:flutter/material.dart';
import 'package:memri/configs/routes/route_navigator.dart';
import 'package:memri/constants/app_styles.dart';
import 'package:memri/core/services/analytics_service.dart';
import 'package:memri/providers/importer_provider.dart';
import 'package:memri/providers/navigation_provider.dart';
import 'package:memri/providers/pod_provider.dart';
import 'package:memri/providers/project_provider.dart';
import 'package:memri/utilities/helpers/app_helper.dart';
import 'package:memri/utilities/helpers/responsive_helper.dart';
import 'package:memri/widgets/project/project_progress_steps.dart';
import 'package:memri/widgets/scaffold/workspace_scaffold.dart';
import 'package:provider/provider.dart';

class ProjectsScreen extends StatefulWidget {
  const ProjectsScreen({Key? key}) : super(key: key);

  @override
  State<ProjectsScreen> createState() => _ProjectsScreenState();
}

class _ProjectsScreenState extends State<ProjectsScreen> {
  late final _navigationProvider =
      Provider.of<NavigationProvider>(context, listen: false);
  late final _iProvider = Provider.of<ImporterProvider>(context, listen: false);
  late final _podProvider = Provider.of<PodProvider>(context, listen: false);
  ScrollController _bodyScrollController = ScrollController();
  final ScrollController _supportScrollController = ScrollController();

  @override
  void initState() {
    Provider.of<ProjectProvider>(context, listen: false).getProjectSummaries();
    _iProvider.fetchImportersData();
    if (_podProvider.state == AuthState.authenticated ||
        _podProvider.state == AuthState.savedKeys) {
      _iProvider.fetchActiveImporters();
    }
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return WorkspaceScaffold(
      currentItem: NavigationItem.projects,
      backgroundColor: app.colors.greyBackGround,
      child: Consumer<ProjectProvider>(
        builder: (_, pProvider, __) => Consumer<ImporterProvider>(
          builder: (_, iProvider, __) => ResponsiveHelper(context).isSmallScreen
              ? _buildMobileBody(pProvider, iProvider)
              : _buildDesktopBody(pProvider, iProvider),
        ),
      ),
    );
  }

  Widget _buildDesktopBody(
      ProjectProvider pProvider, ImporterProvider iProvider) {
    return Container(
      margin: EdgeInsets.symmetric(horizontal: 30),
      child: Row(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Expanded(
            flex: ResponsiveHelper(context).isMediumScreen ? 5 : 3,
            child: Container(
              height: pProvider.projects.isEmpty
                  ? MediaQuery.of(context).size.height
                  : null,
              color: app.colors.white,
              child: SingleChildScrollView(
                controller: _bodyScrollController,
                child: _buildBody(pProvider, iProvider),
              ),
            ),
          ),
          SizedBox(width: 2),
          Expanded(
            flex: ResponsiveHelper(context).isMediumScreen ? 3 : 1,
            child: SingleChildScrollView(
              controller: _supportScrollController,
              child: Container(
                color: app.colors.white,
                padding: EdgeInsets.all(20),
                child: _buildSupportSideBar(),
              ),
            ),
          ),
        ],
      ),
    );
  }

  Widget _buildMobileBody(
      ProjectProvider pProvider, ImporterProvider iProvider) {
    _bodyScrollController = ScrollController();
    return SingleChildScrollView(
      controller: _bodyScrollController,
      child: Column(
        children: [
          _buildBody(pProvider, iProvider),
          SizedBox(height: 2),
          Container(
            color: app.colors.white,
            padding: EdgeInsets.all(30),
            child: _buildSupportSideBar(),
          ),
        ],
      ),
    );
  }

  Widget _buildBody(ProjectProvider pProvider, ImporterProvider iProvider) =>
      (!iProvider.dataForLabelingInPod || pProvider.projects.isEmpty)
          ? _buildEmptyState(pProvider)
          : _buildProjectList(pProvider, iProvider);

  Widget _buildSupportSideBar() {
    return Column(
      mainAxisSize: MainAxisSize.min,
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        app.images.signFourth(),
        SizedBox(height: 50),
        Text('Support', style: AppStyles.headline2),
        SizedBox(height: 10),
        Text(
          'Need some help or have questions about setting up a project? Browse our step-by-step guides and documentation or chat with us.',
          style: AppStyles.bodyText1.copyWith(color: app.colors.brandGreyText),
        ),
        SizedBox(height: 25),
        TextButton(
          style: ButtonStyle(
              padding: MaterialStateProperty.all(EdgeInsets.all(0))),
          onPressed: _navigationProvider.openGuideLink,
          child: Row(
            mainAxisSize: MainAxisSize.min,
            children: [
              Text(
                'Getting started',
                style: AppStyles.bodyText1.copyWith(color: app.colors.primary),
              ),
              SizedBox(width: 10),
              app.icons.arrowRight(color: app.colors.primary),
            ],
          ),
        ),
        SizedBox(height: 30),
        TextButton(
          style: ButtonStyle(
              padding: MaterialStateProperty.all(EdgeInsets.all(0))),
          onPressed: _navigationProvider.openDiscordLink,
          child: Row(
            mainAxisSize: MainAxisSize.min,
            children: [
              Text(
                'Chat with us on Discord',
                style: AppStyles.bodyText1.copyWith(color: app.colors.primary),
              ),
              SizedBox(width: 10),
              app.icons.arrowRight(color: app.colors.primary),
            ],
          ),
        ),
        SizedBox(height: 15),
      ],
    );
  }

  Widget _buildProjectList(
      ProjectProvider provider, ImporterProvider iProvider) {
    return Container(
      width: MediaQuery.of(context).size.width,
      color: app.colors.greyBackGround,
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          ...List.generate(
            provider.projects.length,
            (index) {
              var projectName = provider.projects[index].get('name');
              var projectId = provider.projects[index].get('id');
              var summary =
                  provider.datasetSummaries[projectId] ?? DatasetSummary();
              return InkWell(
                onTap: () => RouteNavigator.navigateTo(
                  context: context,
                  route: Routes.projectsSummary,
                  param: {'id': projectId},
                ),
                child: Container(
                  color: app.colors.white,
                  margin: EdgeInsets.only(bottom: 2),
                  child: Stack(
                    children: [
                      Positioned(
                        top: 8,
                        right: 8,
                        child: app.icons.arrowTopRight(
                          color: app.colors.brandOrange,
                          height: 50,
                          width: 50,
                        ),
                      ),
                      Container(
                        padding: EdgeInsets.all(32),
                        child: Column(
                          children: [
                            Row(
                              children: [
                                Text(projectName, style: AppStyles.headline2),
                              ],
                            ),
                            SizedBox(height: 32),
                            Row(
                              crossAxisAlignment: CrossAxisAlignment.start,
                              children: [
                                Column(
                                  crossAxisAlignment: CrossAxisAlignment.start,
                                  children: [
                                    Row(
                                      children: const [
                                        Text('DATA SOURCE',
                                            style: AppStyles.smallCaps),
                                      ],
                                    ),
                                    Row(
                                      children: const [
                                        Text('Whatsapp', style: AppStyles.ul),
                                      ],
                                    ),
                                  ],
                                ),
                                Spacer(),
                                Column(
                                  crossAxisAlignment: CrossAxisAlignment.start,
                                  children: [
                                    Row(
                                      children: const [
                                        Text('TEXT CLASSIFICATION',
                                            style: AppStyles.smallCaps),
                                      ],
                                    ),
                                    Row(
                                      children: [
                                        Text('Submitted',
                                            style: AppStyles.ul.copyWith(
                                                color:
                                                    app.colors.brandGreyText)),
                                        SizedBox(width: 8),
                                        Text('${summary.numLabeled}',
                                            style: AppStyles.ul),
                                        SizedBox(width: 8),
                                        Text('${summary.percentageLabeled}%',
                                            style: AppStyles.ul),
                                      ],
                                    ),
                                    Row(
                                      children: [
                                        Text('Skipped',
                                            style: AppStyles.ul.copyWith(
                                                color:
                                                    app.colors.brandGreyText)),
                                        SizedBox(width: 8),
                                        Text('${summary.numSkipped}',
                                            style: AppStyles.ul),
                                        SizedBox(width: 8),
                                        Text('${summary.percentageSkipped}%',
                                            style: AppStyles.ul),
                                      ],
                                    ),
                                    Row(
                                      children: [
                                        Text('Dataset',
                                            style: AppStyles.ul.copyWith(
                                                color:
                                                    app.colors.brandGreyText)),
                                        SizedBox(width: 8),
                                        Text('${summary.datasetSize}',
                                            style: AppStyles.ul),
                                      ],
                                    ),
                                  ],
                                ),
                                Spacer(),
                                Column(
                                  crossAxisAlignment: CrossAxisAlignment.start,
                                  children: [
                                    Row(
                                      children: const [
                                        Text('MODEL',
                                            style: AppStyles.smallCaps),
                                      ],
                                    ),
                                    Row(
                                      children: [
                                        Text('Version',
                                            style: AppStyles.ul.copyWith(
                                                color:
                                                    app.colors.brandGreyText)),
                                        SizedBox(width: 8),
                                        Text('1.0', style: AppStyles.ul),
                                      ],
                                    ),
                                    Row(
                                      children: [
                                        Text('Dataset',
                                            style: AppStyles.ul.copyWith(
                                                color:
                                                    app.colors.brandGreyText)),
                                        SizedBox(width: 8),
                                        Text('${summary.datasetSize}',
                                            style: AppStyles.ul),
                                      ],
                                    ),
                                  ],
                                ),
                                Spacer(),
                              ],
                            ),
                            SizedBox(height: 4)
                          ],
                        ),
                      )
                    ],
                  ),
                ),
              );
            },
          ),
          // TODO temporary button, should move it to top bar
          Row(
            children: [
              Padding(
                padding: const EdgeInsets.symmetric(vertical: 30),
                child: _buildProjectButton(iProvider),
              ),
              SizedBox(width: 20),
              _buildSampleProjectButton(),
            ],
          ),
        ],
      ),
    );
  }

  Widget _buildEmptyState(ProjectProvider provider) {
    return Container(
      width: MediaQuery.of(context).size.width,
      // height: ResponsiveHelper(context).isSmallScreen
      //     ? null
      //     : MediaQuery.of(context).size.height,
      color: app.colors.white,
      padding: EdgeInsets.all(30),
      child: Consumer<ImporterProvider>(builder: (_, importerProvider, __) {
        return Column(
          crossAxisAlignment: CrossAxisAlignment.start,
          children: [
            if (!ResponsiveHelper(context).isSmallScreen) SizedBox(height: 16),
            Text(
              'Start building your first data app',
              style: AppStyles.headline1,
            ),
            SizedBox(height: 32),
            Text(
              'Memri allows to build, deploy and share ML apps on personal data. \n\nTo build a data app, you’ll need to create a new project.',
              style: AppStyles.bodyText1.copyWith(fontSize: 14),
            ),
            SizedBox(height: 50),
            ProjectProgressSteps(
                step: importerProvider.isAnyImportersActive ? 2 : 1),
            SizedBox(height: 50),
            Row(
              children: [
                _buildProjectButton(importerProvider),
                SizedBox(width: 20),
                _buildSampleProjectButton(),
              ],
            ),
            SizedBox(height: 16),
          ],
        );
      }),
    );
  }

  Widget _buildProjectButton(ImporterProvider importerProvider) {
    return TextButton(
      onPressed: () => RouteNavigator.navigateTo(
          context: context,
          route: importerProvider.isAnyImportersActive
              ? Routes.projectsCreate
              : Routes.data),
      style: primaryButtonStyle,
      child: SizedBox(
        height: 36,
        child: Row(
          mainAxisSize: MainAxisSize.min,
          children: [
            Text(importerProvider.isAnyImportersActive
                ? 'Create your project'
                : 'Connect data'),
            SizedBox(width: 10),
            app.icons.plus(),
          ],
        ),
      ),
    );
  }

  Widget _buildSampleProjectButton() {
    return TextButton(
      onPressed: () {
        AnalyticsService().logImporterSelect('sample_whatsapp');
        RouteNavigator.navigateTo(
            context: context,
            route: Routes.projectsCreate,
            param: {'isSample': true});
      },
      style: secondaryButtonStyle,
      child: SizedBox(
        height: 36,
        child: Row(
          mainAxisSize: MainAxisSize.min,
          children: const [Text('Create project with sample data')],
        ),
      ),
    );
  }
}
